from onegov.core.orm.session_manager import SessionManager
from sqlalchemy import inspect
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy_utils import TranslationHybrid

MISSING = object()


#: The base for all OneGov Core ORM Models
class ModelBase(object):
    def __getstate__(self):
        """ Makes sure the session manager attached to the model is not
        pickled (stored in memcached). Not only does it not make sense to
        have connection state stored, it also leads to errors.

        Cached instances have to be merged back into the session as a result,
        but that's something you have to do anyway when pickling sqlalchemy
        models.

        """
        if 'session_manager' in self.__dict__:
            del self.__dict__['session_manager']

        return self.__dict__.copy()

    def __setstate__(self, state):
        self.__dict__ = state

    @classmethod
    def get_polymorphic_class(cls, identity_value, default=MISSING):
        """ Returns the polymorphic class if it exists, given the value
        of the polymorphic identity.

        Asserts that the identity is actually found, unless a default is
        provided.

        """
        mapper = inspect(cls).polymorphic_map.get(identity_value)

        if default is MISSING:
            assert mapper, "No such polymorphic_identity: {}".format(
                identity_value
            )

        return mapper and mapper.class_ or default

Base = declarative_base(cls=ModelBase)

#: A translation hybrid integrated with OneGov Core. See also:
#: http://sqlalchemy-utils.readthedocs.org/en/latest/internationalization.html
translation_hybrid = TranslationHybrid(
    current_locale=lambda obj: obj.session_manager.current_locale,
    default_locale=lambda obj: obj.session_manager.default_locale,
)


def find_models(base, is_match):
    """ Finds the ORM models in the given ORM base class that match a filter.

    The filter is called with each class in the instance and it is supposed
    to return True if it matches.

    For example, find all SQLAlchemy models that use
    :class:`~onegov.core.orm.mixins.ContentMixin`::

        from onegov.core.orm.mixins import ContentMixin
        find_models(base, is_match=lambda cls: issubclass(cls, ContentMixin))

    """
    for cls in base.__subclasses__():
        if is_match(cls):
            yield cls

        for cls in find_models(cls, is_match):
            yield cls

__all__ = ['Base', 'SessionManager', 'translation_hybrid', 'find_models']



from morepath import Identity
from onegov.core import Framework


class IdentityPolicy(object):
    """ Stores the tokens associated with the identity on the browser session

    """

    required_keys = {'userid', 'role', 'application_id'}

    def identify(self, request):
        try:
            identifiers = {
                key: request.browser_session[key] for key in self.required_keys
            }
        except KeyError:
            return None
        else:
            return Identity(**identifiers)

    def remember(self, response, request, identity):
        for key in self.required_keys:
            request.browser_session[key] = getattr(identity, key)

    def forget(self, response, request):
        for key in self.required_keys:
            if request.browser_session.has(key):
                del request.browser_session[key]


@Framework.identity_policy()
def identity_policy():
    return IdentityPolicy()


@Framework.verify_identity()
def verify_identity(identity):
    # trust the identity established by the identity policy (we could keep
    # checking if the user is really in the database here - or if it was
    # removed in the meantime)
    return True



import click

from contextlib import contextmanager
from datetime import datetime
from sqlalchemy import event
from sqlalchemy.engine import Engine
from sqlparse import format


class Timer(object):
    """ A timer that works just like a stopwatch. """

    def start(self):
        self.started = datetime.utcnow()

    def stop(self):
        return datetime.utcnow() - self.started


def print_query(query):
    """ Pretty prints the given query. """
    formatted = format(query.decode('utf-8'), reindent=True)
    formatted = formatted.replace('\n', '\n  ')

    print('> {}'.format(formatted))


@contextmanager
def analyze_sql_queries(report='summary'):
    """ Analyzes the sql-queries executed during its context. There are three
    levels of information (report argument):

    * 'summary' (only show the number of queries)
    * 'redundant' (show summary and the actual redundant queries)
    * 'all' (show summary and all executed queries)

    Use this with a with_statement::

        with analyze_sql_queries():
            ... # <- analyzes all sql queries that happen inside here

    """

    assert report in {'summary', 'redundant', 'all'}

    queries = {}
    timer = Timer()

    @event.listens_for(Engine, 'before_cursor_execute')
    def before_exec(conn, cursor, statement, parameters, context, executemany):
        timer.start()

    @event.listens_for(Engine, 'after_cursor_execute')
    def after_exec(conn, cursor, statement, parameters, context, executemany):
        runtime = timer.stop()

        def handle_query(query):
            if report == 'all':
                print_query(query)

            if query not in queries:
                queries[query] = 1
            else:
                queries[query] += 1

        if isinstance(parameters, dict):
            handle_query(cursor.mogrify(statement, parameters))
        else:
            for parameter in parameters:
                handle_query(cursor.mogrify(statement, parameter))

        if report == 'all':
            print('< took {}'.format(runtime))
    yield

    event.remove(Engine, 'before_cursor_execute', before_exec)
    event.remove(Engine, 'after_cursor_execute', after_exec)

    total_queries = sum(queries.values())
    redundant_queries = sum(1 for v in queries.values() if v > 1)

    if total_queries > 10:
        total_queries = click.style(str(total_queries), 'red')
    elif total_queries > 5:
        total_queries = click.style(str(total_queries), 'yellow')
    else:
        total_queries = click.style(str(total_queries), 'green')

    if redundant_queries:
        redundant_queries = click.style(str(redundant_queries), 'red')

    if total_queries != '0':
        print("executed {} queries, {} of which were redundant".format(
            total_queries, redundant_queries))

    if redundant_queries and report == 'redundant':
        print("The following queries were redundant:")
        for query, count in queries.items():
            if count > 1:
                print_query(query)



from chameleon import PageTemplate
from onegov.core import cache
from onegov.core import Framework
from webtest import TestApp as Client


CALL_COUNT = 0


def test_cache_connections():

    class App(Framework):
        pass

    @App.path(path='/')
    class Root(object):
        pass

    @App.html(model=Root)
    def view_root(self, request):
        @request.app.cache.cache_on_arguments()
        def call_count():
            global CALL_COUNT
            CALL_COUNT += 1
            return str(CALL_COUNT)

        return call_count()

    app = App()
    app.namespace = 'towns'
    app.configure_application(cache_connections=2, disable_memcached=True)

    client = Client(app)

    # with a limit of two connections we'll see that whenever a third
    # applicaiton is used, the last used application is lost
    #
    # we won't necessarily see this with memcached - memcached does it's own
    # lru and we don't try to influence that - here we lose the cache because
    # if we use a memory cache backend we lose the cache when discarding
    # the backend
    app.set_application_id('towns/washington')
    assert client.get('/').text == '1'
    assert client.get('/').text == '1'

    app.set_application_id('towns/arlington')
    assert client.get('/').text == '2'
    assert client.get('/').text == '2'

    app.set_application_id('towns/washington')
    assert client.get('/').text == '1'

    app.set_application_id('towns/detroit')
    assert client.get('/').text == '3'

    app.set_application_id('towns/washington')
    assert client.get('/').text == '1'

    app.set_application_id('towns/detroit')
    assert client.get('/').text == '3'

    app.set_application_id('towns/arlington')
    assert client.get('/').text == '4'


def test_unreachable_backend_proxy():

    region = cache.create_backend('ns', 'dogpile.cache.memcached', arguments={
        'url': '127.0.0.1:12345'
    })

    region.delete('test')
    region.set('test', 'value')
    assert region.get('test') is cache.NO_VALUE

    region.delete_multi(['foo', 'bar'])
    region.set_multi({
        'foo': 1, 'bar': 2
    })
    assert region.get_multi(['foo', 'bar']) == [cache.NO_VALUE, cache.NO_VALUE]


def test_cache_key():
    region = cache.create_backend('ns', 'dogpile.cache.memcached', arguments={
        'url': '127.0.0.1:12345'
    })

    region.set('x' * 500, 'y')


def test_cache_page_template():
    region = cache.create_backend('ns', 'onegov.core.memcached', arguments={
        'url': '127.0.0.1:12345'
    })

    region.set('template', PageTemplate('<!DOCTYPE html>'))
    region.get('template')



from onegov.core.orm.mixins.content import (
    ContentMixin,
    meta_property,
    content_property
)
from onegov.core.orm.mixins.timestamp import TimestampMixin

__all__ = [
    'content_property',
    'meta_property',
    'ContentMixin',
    'TimestampMixin'
]



import morepath
import os.path
import time

from onegov.core import Framework
from onegov.core import utils
from onegov.core.static import StaticFile
from webtest import TestApp as Client


def test_static_file(temporary_directory):

    class App(Framework):
        serve_static_files = True

    @App.static_directory()
    def get_static_files():
        return temporary_directory

    morepath.commit(App)

    app = App()
    app.configure_application()
    app.namespace = 'test'
    app.set_application_id('test/test')

    with open(os.path.join(temporary_directory, 'robots.txt'), 'w') as f:
        f.write('foobar')

    app.serve_static_files = False
    assert StaticFile.from_application(app, 'robots.txt') is None

    app.serve_static_files = True
    assert StaticFile.from_application(app, 'robots.txt').path == 'robots.txt'
    assert StaticFile.from_application(app, '../robots.txt') is None


def test_static_file_app(temporary_directory):

    class App(Framework):
        serve_static_files = True

    @App.static_directory()
    def get_static_files():
        return temporary_directory

    utils.scan_morepath_modules(App)
    morepath.commit(App)

    with open(os.path.join(temporary_directory, 'robots.txt'), 'w') as f:
        f.write('foobar')

    app = App()
    app.configure_application()
    app.namespace = 'test'
    app.set_application_id('test/test')

    c = Client(app)
    assert c.get('/static/robots.txt').text == 'foobar'

    # make sure the if-modified-since header is honored
    timestamp = time.gmtime(time.time() + 1)
    modified = time.strftime('%a, %d %b %Y %H:%M:%S GMT', timestamp)
    headers = {'If-Modified-Since': modified}

    response = c.get('/static/robots.txt', headers=headers, expect_errors=True)
    assert response.status_code == 304

    # make sure inexistant files return a 404
    assert c.get('/static/humans.txt', expect_errors=True).status_code == 404


def test_root_file_app(temporary_directory):
    class App(Framework):
        serve_static_files = True

    @App.static_directory()
    def get_static_files():
        return temporary_directory

    class RobotsTxt(StaticFile):
        pass

    @App.path(model=RobotsTxt, path='robots.txt')
    def get_favicon(app, absorb):
        return StaticFile.from_application(app, 'robots.txt')

    utils.scan_morepath_modules(App)
    morepath.commit(App)

    with open(os.path.join(temporary_directory, 'robots.txt'), 'w') as f:
        f.write('foobar')

    app = App()
    app.configure_application()
    app.namespace = 'test'
    app.set_application_id('test/test')

    c = Client(app)
    assert c.get('/robots.txt').text == 'foobar'


def test_static_files_directive(temporary_directory):

    a = os.path.join(temporary_directory, 'a')
    b = os.path.join(temporary_directory, 'b')

    for path in (a, b):
        os.mkdir(path)

        with open(os.path.join(path, 'foobar.txt'), 'w') as f:
            f.write(path)

    class A(Framework):
        serve_static_files = True

    class B(A):
        serve_static_files = True

    @A.static_directory()
    def get_static_files_a():
        return a

    @B.static_directory()
    def get_static_files_b():
        return b

    utils.scan_morepath_modules(A)
    utils.scan_morepath_modules(B)

    morepath.commit(A)
    morepath.commit(B)

    app_a = A()
    app_b = B()

    for app in (app_a, app_b):
        app.configure_application()
        app.namespace = 'test'
        app.set_application_id('test/test')

    assert app_a.static_files == [a]
    assert app_b.static_files == [b, a]

    assert Client(app_a).get('/static/foobar.txt').text == a
    assert Client(app_b).get('/static/foobar.txt').text == b



from datetime import datetime
from onegov.core.custom_json import dumps, loads


def test_custom_json():

    dt = datetime(2015, 6, 25, 12, 0)

    data = {
        'datetime': dt,
        'date': dt.date(),
        'time': dt.time(),
        'generator': (x for x in range(0, 4))
    }

    text = dumps(data)

    assert '__time__@12:00:00' in text
    assert '__date__@2015-06-25' in text
    assert '__datetime__@2015-06-25T12:00:00' in text
    assert '[0, 1, 2, 3]' in text

    data['generator'] = [x for x in range(0, 4)]
    assert loads(text) == data



import morepath
import pytest
import pickle
import sqlalchemy
import time
import transaction
import uuid

from datetime import datetime
from onegov.core.framework import Framework
from onegov.core.orm import (
    ModelBase, SessionManager, translation_hybrid, find_models
)
from onegov.core.orm.abstract import AdjacencyList
from onegov.core.orm.mixins import (
    meta_property, content_property, ContentMixin, TimestampMixin
)
from onegov.core.orm.types import HSTORE, JSON, UTCDateTime, UUID
from onegov.core.security import Private
from onegov.core.utils import scan_morepath_modules
from psycopg2.extensions import TransactionRollbackError
from pytz import timezone
from sqlalchemy import Column, Integer, Text
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.mutable import MutableDict
from threading import Thread
from webob.exc import HTTPUnauthorized, HTTPConflict
from webtest import TestApp as Client


class PicklePage(AdjacencyList):
    __tablename__ = 'picklepages'


def test_is_valid_schema(postgres_dsn):
    mgr = SessionManager(postgres_dsn, None)
    assert not mgr.is_valid_schema('pg_test')
    assert not mgr.is_valid_schema('-- or 1=1')
    assert not mgr.is_valid_schema('0')
    assert not mgr.is_valid_schema('information_schema')
    assert not mgr.is_valid_schema('public')
    assert not mgr.is_valid_schema('my--schema')
    assert mgr.is_valid_schema('my_schema')
    assert mgr.is_valid_schema('my-schema')


def test_independent_sessions(postgres_dsn):
    Base = declarative_base()

    class Document(Base):
        __tablename__ = 'document'
        id = Column(Integer, primary_key=True)

    mgr = SessionManager(postgres_dsn, Base)

    mgr.set_current_schema('foo')
    s1 = mgr.session()

    s1.add(Document())
    s1.flush()

    mgr.set_current_schema('bar')
    s2 = mgr.session()

    assert s1 is not s2

    assert s1.query(Document).count() == 1
    assert s2.query(Document).count() == 0

    mgr.dispose()


def test_independent_managers(postgres_dsn):
    Base = declarative_base()

    class Document(Base):
        __tablename__ = 'document'
        id = Column(Integer, primary_key=True)

    one = SessionManager(postgres_dsn, Base)
    two = SessionManager(postgres_dsn, Base)

    one.set_current_schema('foo')
    two.set_current_schema('foo')

    assert one.session() is not two.session()
    assert one.session() is one.session()
    assert two.session() is two.session()

    one.session().add(Document())
    one.session().flush()

    assert one.session().query(Document).count() == 1
    assert two.session().query(Document).count() == 0

    one.set_current_schema('bar')
    one.session().info == {'schema': 'bar'}
    two.session().info == {'schema': 'foo'}

    one.dispose()
    two.dispose()


def test_create_schema(postgres_dsn):
    Base = declarative_base()

    class Document(Base):
        __tablename__ = 'document'

        id = Column(Integer, primary_key=True)
        title = Column(Text)

    mgr = SessionManager(postgres_dsn, Base)

    # we need a schema to use the session manager and it can't be 'public'
    mgr.set_current_schema('testing')

    def existing_schemas():
        # DO NOT copy this query, it's insecure (which is fine in testing)
        return set(
            r['schema_name'] for r in mgr.engine.execute(
                'SELECT schema_name FROM information_schema.schemata'
            )
        )

    def schema_tables(schema):
        # DO NOT copy this query, it's insecure (which is fine in testing)
        return set(
            r['tablename'] for r in mgr.engine.execute((
                "SELECT tablename FROM pg_catalog.pg_tables "
                "WHERE schemaname = '{}'".format(schema)
            ))
        )

    assert 'testing' in existing_schemas()
    assert 'new' not in existing_schemas()

    mgr.ensure_schema_exists('new')

    assert 'new' in existing_schemas()
    assert 'document' in schema_tables('new')

    mgr.dispose()


def test_schema_bound_session(postgres_dsn):
    Base = declarative_base()

    class Document(Base):
        __tablename__ = 'documents'

        id = Column(Integer, primary_key=True)
        title = Column(Text)

    mgr = SessionManager(postgres_dsn, Base)
    mgr.set_current_schema('foo')
    session = mgr.session()

    session.add(Document(title='Welcome to Foo'))
    transaction.commit()

    assert session.query(Document).one().title == 'Welcome to Foo'

    mgr.set_current_schema('bar')
    session = mgr.session()

    assert session.query(Document).first() is None

    mgr.set_current_schema('foo')
    session = mgr.session()

    assert session.query(Document).one().title == 'Welcome to Foo'

    mgr.dispose()


def test_session_scope(postgres_dsn):
    Base = declarative_base()

    mgr = SessionManager(postgres_dsn, Base)

    mgr.set_current_schema('foo')
    foo_session = mgr.session()

    mgr.set_current_schema('bar')
    bar_session = mgr.session()

    assert foo_session is not bar_session

    mgr.set_current_schema('foo')
    foo_session_2 = mgr.session()

    mgr.set_current_schema('bar')
    bar_session_2 = mgr.session()

    assert foo_session is foo_session_2
    assert bar_session is bar_session_2

    mgr.dispose()


def test_orm_scenario(postgres_dsn):
    # test a somewhat complete ORM scenario in which create and read data
    # for different applications
    Base = declarative_base()

    class App(Framework):
        pass

    class Document(Base):
        __tablename__ = 'documents'

        id = Column(Integer, primary_key=True)
        title = Column(Text, nullable=False)

    class DocumentCollection(object):

        def __init__(self, session):
            self.session = session

        def query(self):
            return self.session.query(Document)

        def all(self):
            return self.query().all()

        def get(self, id):
            return self.query().filter(Document.id == id).first()

        def add(self, title):
            document = Document(title=title)
            self.session.add(document)
            self.session.flush()

            return document

    @App.path(model=DocumentCollection, path='documents')
    def get_documents(app):
        return DocumentCollection(app.session())

    @App.json(model=DocumentCollection)
    def documents_default(self, request):
        return {d.id: d.title for d in self.all()}

    @App.json(model=DocumentCollection, name='add', request_method='POST')
    def documents_add(self, request):
        self.add(title=request.params.get('title'))

    @App.json(model=DocumentCollection, name='error')
    def documents_error(self, request):
        # tries to create a document that should not be created since the
        # request as a whole fails
        self.add('error')

        raise HTTPUnauthorized()

    # this is required for the transactions to actually work, usually this
    # would be onegov.server's job
    scan_morepath_modules(App)

    app = App()
    app.configure_application(dsn=postgres_dsn, base=Base)
    app.namespace = 'municipalities'

    c = Client(app)

    # let's try to add some documents to new york
    app.set_application_id('municipalities/new-york')

    assert c.get('/documents').json == {}
    c.post('/documents/add', {'title': 'Welcome to the big apple!'})

    assert c.get('/documents').json == {
        '1': "Welcome to the big apple!"
    }

    # after that, we switch to chicago, where a different set of documents
    # should exist
    app.set_application_id('municipalities/chicago')

    assert c.get('/documents').json == {}
    c.post('/documents/add', {'title': 'Welcome to the windy city!'})

    assert c.get('/documents').json == {
        '1': "Welcome to the windy city!"
    }

    # finally, let's see if the transaction is rolled back if there's an
    # error during the course of the request
    c.get('/documents/error', expect_errors=True)
    assert c.get('/documents').json == {
        '1': "Welcome to the windy city!"
    }

    app.session_manager.dispose()


def test_i18n_with_request(postgres_dsn):
    Base = declarative_base()

    class App(Framework):
        pass

    class Document(Base):
        __tablename__ = 'documents'

        id = Column(Integer, primary_key=True)

        title_translations = Column(HSTORE, nullable=False)
        title = translation_hybrid(title_translations)

    @App.path(model=Document, path='document')
    def get_document(app):
        return app.session().query(Document).first() or Document(id=1)

    @App.json(model=Document)
    def view_document(self, request):
        return {'title': self.title}

    @App.json(model=Document, request_method='PUT')
    def put_document(self, request):
        self.title = request.params.get('title')
        app.session().merge(self)

    @App.setting(section='i18n', name='default_locale')
    def get_i18n_default_locale():
        return 'de_CH'

    scan_morepath_modules(App)

    app = App()
    app.configure_application(dsn=postgres_dsn, base=Base)
    app.namespace = 'municipalities'
    app.set_application_id('municipalities/new-york')
    app.locales = ['de_CH', 'en_US']

    c = Client(app)
    c.put('/document?title=Dokument')
    assert c.get('/document').json == {'title': 'Dokument'}

    c.set_cookie('locale', 'en_US')
    c.put('/document?title=Document')
    assert c.get('/document').json == {'title': 'Document'}

    c.set_cookie('locale', '')
    assert c.get('/document').json == {'title': 'Dokument'}

    app.session_manager.dispose()


def test_json_type(postgres_dsn):
    Base = declarative_base()

    class Test(Base):
        __tablename__ = 'test'

        id = Column(Integer, primary_key=True)
        data = Column(JSON, nullable=True)

    mgr = SessionManager(postgres_dsn, Base)
    mgr.set_current_schema('testing')

    session = mgr.session()

    test = Test(id=1, data=None)
    session.add(test)
    transaction.commit()

    assert session.query(Test).filter(Test.id == 1).one().data is None

    test = Test(id=2, data={'foo': 'bar'})
    session.add(test)
    transaction.commit()

    assert session.query(Test).filter(Test.id == 2).one().data == {
        'foo': 'bar'
    }

    test = session.query(Test).filter(Test.id == 2).one()
    test.data['foo'] = 'rab'
    transaction.commit()

    assert session.query(Test).filter(Test.id == 2).one().data == {
        'foo': 'rab'
    }

    mgr.dispose()


def test_session_manager_sharing(postgres_dsn):
    Base = declarative_base()

    class Test(Base):
        __tablename__ = 'test'
        id = Column(Integer, primary_key=True)

    mgr = SessionManager(postgres_dsn, Base)
    mgr.set_current_schema('testing')

    test = Test(id=1)

    # session_manager is a weakref proxy so we need to go through some hoops
    # to get the actual instance for a proper identity test
    assert test.session_manager.__repr__.__self__ is mgr

    session = mgr.session()
    session.add(test)
    transaction.commit()

    assert session.query(Test).one().session_manager.__repr__.__self__ is mgr
    mgr.dispose()


def test_session_manager_i18n(postgres_dsn):
    Base = declarative_base()

    class Test(Base):
        __tablename__ = 'test'
        id = Column(Integer, primary_key=True)

        text_translations = Column(HSTORE)
        text = translation_hybrid(text_translations)

    mgr = SessionManager(postgres_dsn, Base)
    mgr.set_current_schema('testing')
    mgr.set_locale(default_locale='en_us', current_locale='en_us')

    test = Test(id=1, text='no')
    assert test.text == 'no'

    mgr.set_locale(default_locale='en_us', current_locale='de_ch')
    assert test.text == 'no'

    test.text_translations['de_ch'] = 'nein'
    assert test.text == 'nein'

    mgr.set_locale(default_locale='en_us', current_locale='en_us')
    assert test.text == 'no'

    session = mgr.session()
    session.add(test)
    transaction.commit()

    test = session.query(Test).one()
    assert test.text == 'no'

    mgr.set_locale(default_locale='en_us', current_locale='de_ch')
    assert test.text == 'nein'

    # make sure the locale is shared with the query as well
    assert mgr.session().query(Test).order_by(Test.text).first()

    mgr.dispose()


def test_uuid_type(postgres_dsn):
    Base = declarative_base()

    class Test(Base):
        __tablename__ = 'test'

        id = Column(UUID, primary_key=True, default=uuid.uuid4)

    mgr = SessionManager(postgres_dsn, Base)
    mgr.set_current_schema('testing')

    session = mgr.session()

    test = Test()
    session.add(test)
    transaction.commit()

    assert isinstance(session.query(Test).one().id, uuid.UUID)

    mgr.dispose()


def test_utc_datetime_naive(postgres_dsn):
    Base = declarative_base()

    class Test(Base):
        __tablename__ = 'test'

        id = Column(Integer, primary_key=True)
        date = Column(UTCDateTime)

    mgr = SessionManager(postgres_dsn, Base)
    mgr.set_current_schema('testing')

    session = mgr.session()

    with pytest.raises(sqlalchemy.exc.StatementError):
        test = Test(date=datetime.now())
        session.add(test)
        session.flush()

    mgr.dispose()


def test_utc_datetime_aware(postgres_dsn):
    Base = declarative_base()

    class Test(Base):
        __tablename__ = 'test'

        id = Column(Integer, primary_key=True)
        date = Column(UTCDateTime)

    mgr = SessionManager(postgres_dsn, Base)
    mgr.set_current_schema('testing')

    session = mgr.session()

    tz = timezone('Europe/Zurich')
    date = datetime(2015, 3, 5, 12, 0).replace(tzinfo=tz)
    test = Test(date=date)
    session.add(test)
    session.flush()
    transaction.commit()

    assert session.query(Test).one().date == date

    mgr.dispose()


def test_timestamp_mixin(postgres_dsn):
    Base = declarative_base()

    class Test(Base, TimestampMixin):
        __tablename__ = 'test'

        id = Column(Integer, primary_key=True)

    mgr = SessionManager(postgres_dsn, Base)
    mgr.set_current_schema('testing')

    session = mgr.session()

    test = Test()
    session.add(test)
    session.flush()
    transaction.commit()

    now = datetime.utcnow()

    assert session.query(Test).one().created.year == now.year
    assert session.query(Test).one().created.month == now.month
    assert session.query(Test).one().created.day == now.day

    assert session.query(Test).one().modified is None

    test = session.query(Test).one()
    test.id = 2
    session.flush()

    assert session.query(Test).one().modified.year == now.year
    assert session.query(Test).one().modified.month == now.month
    assert session.query(Test).one().modified.day == now.day

    mgr.dispose()


def test_content_mixin(postgres_dsn):
    Base = declarative_base()

    class Test(Base, ContentMixin):
        __tablename__ = 'test'

        id = Column(Integer, primary_key=True)

    mgr = SessionManager(postgres_dsn, Base)
    mgr.set_current_schema('testing')

    session = mgr.session()

    test = Test(meta={'filename': 'rtfm'}, content={'text': 'RTFM'})
    session.add(test)
    session.flush()
    transaction.commit()

    session.query(Test).one().meta == {'filename': 'rtfm'}
    session.query(Test).one().content == {'text': 'RTFM'}

    mgr.dispose()


def test_extensions_schema(postgres_dsn):
    Base = declarative_base()

    class Data(Base):
        __tablename__ = 'data'

        id = Column(Integer, primary_key=True)
        data = Column(MutableDict.as_mutable(HSTORE))

    mgr = SessionManager(postgres_dsn, Base)

    for ix, schema in enumerate(('foo', 'bar')):
        mgr.set_current_schema(schema)

        obj = Data()
        obj.data = {}
        obj.data['index'] = str(ix)
        obj.data['schema'] = schema

        mgr.session().add(obj)
        mgr.session().flush()

        transaction.commit()

    for ix, schema in enumerate(('foo', 'bar')):
        mgr.set_current_schema(schema)

        obj = mgr.session().query(Data).one()
        assert obj.data['index'] == str(ix)
        assert obj.data['schema'] == schema

    assert mgr.created_extensions == {'hstore'}


def test_serialization_failure(postgres_dsn):

    Base = declarative_base()

    class Data(Base):
        __tablename__ = 'data'
        id = Column(Integer, primary_key=True)

    class MayFailThread(Thread):

        def __init__(self, dsn, base, schema):
            Thread.__init__(self)
            self.dsn = dsn
            self.base = base
            self.schema = schema

        def run(self):
            mgr = SessionManager(self.dsn, self.base)
            mgr.set_current_schema(self.schema)

            session = mgr.session()

            session.add(Data())
            session.query(Data).delete('fetch')

            time.sleep(0.1)

            try:
                transaction.commit()
            except Exception as e:
                self.exception = e
            else:
                self.exception = None

    # make sure the schema exists already, its creation is not threadsafe
    SessionManager(postgres_dsn, Base).set_current_schema('foo')

    threads = [
        MayFailThread(postgres_dsn, Base, 'foo'),
        MayFailThread(postgres_dsn, Base, 'foo')
    ]

    [t.start() for t in threads]
    [t.join() for t in threads]

    exceptions = [t.exception for t in threads]

    # one will have failed with a rollback error
    rollbacks = [e for e in exceptions if e]
    assert len(rollbacks) == 1
    assert isinstance(rollbacks[0].orig, TransactionRollbackError)


@pytest.mark.parametrize("number_of_retries", range(1, 10))
def test_application_retries(postgres_dsn, number_of_retries):

    Base = declarative_base()

    class App(Framework):
        pass

    @App.path(path='/foo/{id}/{uid}')
    class Document(object):
        def __init__(self, id, uid):
            self.id = id
            self.uid = uid

    @App.path(path='/')
    class Root(object):
        pass

    class Record(Base):
        __tablename__ = 'records'
        id = Column(Integer, primary_key=True)

    @App.view(model=Root, name='init')
    def init_schema(self, request):
        pass  # the schema is initialized by the application

    @App.view(model=Root, name='login')
    def login(self, request):
        @request.after
        def remember(response):
            identity = morepath.Identity(
                userid='admin',
                role='editor',
                application_id=request.app.application_id
            )

            request.app.remember_identity(response, request, identity)

    @App.view(model=Document, permission=Private)
    def provoke_serialization_failure(self, request):
        session = request.app.session()
        session.add(Record())
        session.query(Record).delete('fetch')

        time.sleep(0.2)

    @App.view(model=OperationalError)
    def operational_error_handler(self, request):
        if not hasattr(self, 'orig'):
            return

        if not isinstance(self.orig, TransactionRollbackError):
            return

        raise HTTPConflict()

    @App.setting(section='transaction', name='attempts')
    def get_retry_attempts():
        return number_of_retries

    scan_morepath_modules(App)

    app = App()
    app.configure_application(
        dsn=postgres_dsn,
        base=Base,
        identity_secure=False,
        disable_memcached=True
    )
    app.namespace = 'municipalities'

    # make sure the schema exists already
    app.set_application_id('municipalities/new-york')
    Client(app).get('/init')

    class RequestThread(Thread):
        def __init__(self, app, path):
            Thread.__init__(self)
            self.app = app
            self.path = path

        def run(self):
            try:
                client = Client(self.app)
                client.get('/login')
                self.response = client.get(self.path, expect_errors=True)
            except Exception as e:
                self.exception = e
            else:
                self.exception = None

    # the number of threads that succeed are the number of retries - 1
    threads = [
        RequestThread(app, '/foo/bar/baz')
        for i in range(number_of_retries + 1)
    ]

    [t.start() for t in threads]
    [t.join() for t in threads]

    # no exceptions should happen, we want proper http return codes
    assert len([t.exception for t in threads if t.exception]) == 0
    assert len([t.response for t in threads if t.response]) == len(threads)

    # all responses should be okay, but for one which gets a 409 Conflict
    status_codes = [t.response.status_code for t in threads]
    assert sum(1 for c in status_codes if c == 200) == len(threads) - 1
    assert sum(1 for c in status_codes if c == 409) == 1


def test_orm_signals_independence(postgres_dsn):
    Base = declarative_base()

    class Document(Base):
        __tablename__ = 'documents'
        id = Column(Integer, primary_key=True)

    m1 = SessionManager(postgres_dsn, Base)
    m2 = SessionManager(postgres_dsn, Base)

    m1.set_current_schema('foo')
    m2.set_current_schema('foo')

    m1_inserted, m2_inserted = [], []

    @m1.on_insert.connect
    def on_m1_insert(schema, obj):
        m1_inserted.append((obj, schema))

    @m2.on_insert.connect
    def on_m2_insert(schema, obj):
        m2_inserted.append((obj, schema))

    m1.session().add(Document())
    m1.session().flush()

    assert len(m1_inserted) == 1
    assert len(m2_inserted) == 0

    m2.session().add(Document())
    m2.session().flush()

    assert len(m1_inserted) == 1
    assert len(m2_inserted) == 1


def test_orm_signals_schema(postgres_dsn):
    Base = declarative_base()

    class Document(Base):
        __tablename__ = 'documents'
        id = Column(Integer, primary_key=True)

    mgr = SessionManager(postgres_dsn, Base)

    inserted = []

    @mgr.on_insert.connect
    def on_insert(schema, obj):
        inserted.append((obj, schema))

    mgr.set_current_schema('foo')
    session = mgr.session()
    session.add(Document())
    session.flush()

    assert inserted[0][1] == 'foo'

    mgr.set_current_schema('bar')
    session.add(Document())
    session.flush()

    assert inserted[1][1] == 'bar'


def test_scoped_signals(postgres_dsn):
    Base = declarative_base()

    class Document(Base):
        __tablename__ = 'documents'
        id = Column(Integer, primary_key=True)

    mgr = SessionManager(postgres_dsn, Base)

    inserted = []

    @mgr.on_insert.connect_via('bar')
    def on_insert(schema, obj):
        inserted.append((obj, schema))

    mgr.set_current_schema('foo')
    mgr.session().add(Document())
    mgr.session().flush()

    assert len(inserted) == 0

    mgr.set_current_schema('bar')
    mgr.session().add(Document())
    mgr.session().flush()

    assert len(inserted) == 1


def test_orm_signals_data_flushed(postgres_dsn):
    Base = declarative_base()

    class Document(Base):
        __tablename__ = 'documents'
        id = Column(Integer, primary_key=True)
        body = Column(Text, nullable=True, default=lambda: 'asdf')

    mgr = SessionManager(postgres_dsn, Base)
    mgr.set_current_schema('foo')

    inserted = []

    @mgr.on_insert.connect
    def on_insert(schema, obj):
        inserted.append((obj, schema))

    mgr.session().add(Document())
    mgr.session().flush()

    assert inserted[0][0].id > 0
    assert inserted[0][0].body == 'asdf'


def test_pickle_model(postgres_dsn):

    # pickling doesn't work with inline classes, so we need to use the
    # PicklePage class defined in thos module
    from onegov.core.orm import Base
    mgr = SessionManager(postgres_dsn, Base)
    mgr.set_current_schema('foo')

    # pickling will fail if the session_manager is still attached
    page = PicklePage(name='foobar', title='Foobar')
    assert page.session_manager.__repr__.__self__ is mgr

    # this is why we automatically remove it internally when we pickle
    page = pickle.loads(pickle.dumps(page))

    assert page.name == 'foobar'
    assert page.title == 'Foobar'

    # loading the pickle dump will show that we have lost the session_manager
    assert not hasattr(page, 'session_manager')

    # to get it back we need to merge the object back into the sesssion,
    # which is standard practice for sqlalchemy objects stored in a cache
    page = mgr.session().merge(page)
    assert page.session_manager.__repr__.__self__ is mgr


def test_orm_signals(postgres_dsn):
    Base = declarative_base()

    class Document(Base):
        __tablename__ = 'documents'
        id = Column(Integer, primary_key=True)
        body = Column(Text, nullable=True)

    class Comment(Base):
        __tablename__ = 'comments'
        id = Column(Integer, primary_key=True)
        document_id = Column(Integer, primary_key=True)
        body = Column(Text, nullable=True)

    mgr = SessionManager(postgres_dsn, Base)

    inserted, updated, deleted = [], [], []

    @mgr.on_insert.connect
    def on_insert(schema, obj):
        inserted.append((obj, schema))

    @mgr.on_update.connect
    def on_update(schema, obj):
        updated.append((obj, schema))

    @mgr.on_delete.connect
    def on_delete(schema, obj):
        deleted.append((obj, schema))

    mgr.set_current_schema('foo')
    session = mgr.session()

    # test on_insert
    session.add(Document(id=1))
    session.add(Comment(id=1, document_id=1))
    assert len(inserted) == 0

    session.flush()
    assert len(inserted) == 2

    # the event order is random
    doc = hasattr(inserted[0][0], 'document_id') and inserted[1] or inserted[0]
    com = hasattr(inserted[0][0], 'document_id') and inserted[0] or inserted[1]

    assert doc[0].id == 1
    assert doc[1] == 'foo'

    assert com[0].id == 1
    assert com[0].document_id == 1
    assert com[1] == 'foo'

    transaction.commit()
    assert len(inserted) == 2

    # test on_update
    document = session.query(Document).filter(Document.id == 1).one()
    document.id = 3
    assert len(updated) == 0

    session.flush()
    assert len(updated) == 1
    assert updated[0][0].id == 3
    assert updated[0][1] == 'foo'

    transaction.commit()
    assert len(updated) == 1

    # test on_delete
    session.delete(document)
    assert len(deleted) == 0

    session.flush()
    assert len(deleted) == 1

    session.flush()
    assert len(deleted) == 1
    assert deleted[0][0].id == 3
    assert deleted[0][1] == 'foo'

    # test on_update with bulk update
    del inserted[:]
    del updated[:]
    del deleted[:]

    session.add(Document(id=1))
    session.add(Document(id=2))
    session.flush()

    assert len(inserted) == 2

    session.query(Document).update({'body': 'hello world'})
    assert len(updated) == 2
    assert updated[0][0].body == 'hello world'
    assert updated[0][1] == 'foo'
    assert updated[1][0].body == 'hello world'
    assert updated[1][1] == 'foo'

    transaction.commit()

    # test on_delete with bulk delete
    session.add(Comment(id=1, document_id=2, body='foo'))
    session.add(Comment(id=2, document_id=2, body='bar'))
    transaction.commit()
    session.query(Comment).filter(Comment.document_id == 2).delete()
    assert len(deleted) == 2
    assert deleted[0][0].id == 1
    assert deleted[0][0].document_id == 2
    assert deleted[0][1] == 'foo'
    assert deleted[1][0].id == 2
    assert deleted[1][0].document_id == 2
    assert deleted[1][1] == 'foo'

    # .. since those objects are never loaded, the body is not present
    assert not deleted[0][0].body
    assert not deleted[1][0].body


def test_get_polymorphic_class():
    Base = declarative_base(cls=ModelBase)

    class Plain(Base):
        __tablename__ = 'plain'
        id = Column(Integer, primary_key=True)

    class Base(Base):
        __tablename__ = 'polymorphic'

        id = Column(Integer, primary_key=True)
        type = Column(Text)

        __mapper_args__ = {
            'polymorphic_on': 'type'
        }

    class ChildA(Base):
        __mapper_args__ = {'polymorphic_identity': 'A'}

    class ChildB(Base):
        __mapper_args__ = {'polymorphic_identity': 'B'}

    assert Plain.get_polymorphic_class('A', None) is None
    assert Plain.get_polymorphic_class('B', None) is None
    assert Plain.get_polymorphic_class('C', None) is None

    assert Plain.get_polymorphic_class('A', 1) == 1
    assert Plain.get_polymorphic_class('B', 2) == 2
    assert Plain.get_polymorphic_class('C', 3) == 3

    assert Base.get_polymorphic_class('A') is ChildA
    assert ChildA.get_polymorphic_class('A') is ChildA
    assert ChildB.get_polymorphic_class('A') is ChildA

    assert Base.get_polymorphic_class('B') is ChildB
    assert ChildA.get_polymorphic_class('B') is ChildB
    assert ChildB.get_polymorphic_class('B') is ChildB

    assert Base.get_polymorphic_class('C', None) is None
    assert ChildA.get_polymorphic_class('C', None) is None
    assert ChildB.get_polymorphic_class('C', None) is None

    with pytest.raises(AssertionError) as assertion_info:
        Base.get_polymorphic_class('C')

    assert "No such polymorphic_identity: C" in str(assertion_info.value)


def test_content_properties():

    class Content(object):
        meta = {}
        content = {}

        type = meta_property('type')
        name = content_property('name')

        @name.setter
        def name(self, value):
            self.content['name'] = value
            self.content['name2'] = value

        @name.deleter
        def name(self):
            del self.content['name']
            del self.content['name2']

    content = Content()
    assert content.type is None
    assert content.name is None

    content.type = 'page'
    assert content.type == 'page'
    assert content.meta['type'] == 'page'
    del content.type
    assert content.type is None

    content.name = 'foobar'
    assert content.name == 'foobar'
    assert content.content['name'] == 'foobar'
    assert content.content['name2'] == 'foobar'

    del content.name

    assert content.name is None
    assert content.content == {}

    content.meta = None
    assert content.type is None
    content.type = 'Foobar'
    assert content.type == 'Foobar'


def test_find_models():

    Base = declarative_base(cls=ModelBase)

    class Mixin(object):
        pass

    class A(Base):
        __tablename__ = 'plain'
        id = Column(Integer, primary_key=True)

    class B(Base, Mixin):
        __tablename__ = 'polymorphic'
        id = Column(Integer, primary_key=True)

    results = list(find_models(Base, lambda cls: issubclass(cls, Mixin)))
    assert results == [B]

    results = list(find_models(Base, lambda cls: not issubclass(cls, Mixin)))
    assert results == [A]

    results = list(find_models(Base, lambda cls: True))
    assert results == [A, B]



import pytest
import tempfile

from io import BytesIO
from onegov.core import utils
from onegov.core.csv import (
    CSVFile,
    convert_list_of_dicts_to_csv,
    convert_list_of_dicts_to_xlsx,
    convert_xls_to_csv,
    detect_encoding,
    match_headers,
    normalize_header,
    parse_header,
)
from onegov.core.errors import (
    AmbiguousColumnsError,
    DuplicateColumnNamesError,
    EmptyLineInFileError,
    MissingColumnsError,
)
from openpyxl import load_workbook


def test_parse_header():
    assert parse_header("   Firtst name;  LastNAME; Designation")\
        == ['firtst name', 'lastname', 'designation']
    assert parse_header("a") == ['a']
    assert parse_header("") == []
    assert parse_header("a,b,c;d") == ['a', 'b', 'c;d']


def test_normalize_header():
    assert normalize_header("") == ""
    assert normalize_header("Wääh") == "waah"
    assert normalize_header(" a   b\tc ") == "a b c"


def test_detect_encoding():
    assert detect_encoding(BytesIO('jöö'.encode('ISO-8859-2'))) == 'cp1252'
    assert detect_encoding(BytesIO('jöö'.encode('utf-8'))) == 'utf-8'

    assert detect_encoding(
        BytesIO('abcdefabcdefabcdefabcdefö'.encode('utf-8'))
    ) == 'utf-8'


def test_simple_csv_file():
    data = (
        b'Datum, Reale Temperatur, Gef\xfchlte Temperatur\n'
        b'01.01.2015, 5, k\xfchl\n'
        b'02.01.2015, 0, kalt'
    )

    csv = CSVFile(
        BytesIO(data), ['datum', 'reale_temperatur', 'gefuhlte_temperatur']
    )

    assert list(csv.headers.keys()) == [
        'datum', 'reale_temperatur', 'gefuhlte_temperatur'
    ]
    list(csv.lines) == [
        csv.rowtype(
            rownumber=1,
            datum='01.01.2015',
            reale_temperatur='5',
            gefuhlte_temperatur='kühl'
        ),
        csv.rowtype(
            rownumber=2,
            datum='01.01.2015',
            reale_temperatur='5',
            gefuhlte_temperatur='kalt'
        ),
    ]


@pytest.mark.parametrize("excel_file", [
    utils.module_path('onegov.core', 'tests/fixtures/excel.xls'),
    utils.module_path('onegov.core', 'tests/fixtures/excel.xlsx'),
])
def test_convert_to_csv(excel_file):
    with open(excel_file, 'rb') as f:
        convert_xls_to_csv(f, None)
        convert_xls_to_csv(f, '')
        with pytest.raises(IOError):
            convert_xls_to_csv(f, 'Sheet 3')

        csv = CSVFile(convert_xls_to_csv(f), ['ID', 'Namä', 'Date'])

        assert list(csv.headers.keys()) == ['ID', 'Namä', 'Date']
        assert list(csv.lines) == [
            csv.rowtype(
                rownumber=2,
                id='1',
                nama='Döner',
                date='31.12.2015'
            ),
            csv.rowtype(
                rownumber=3,
                id='2',
                nama='“Cola”',
                date='31.12.2014 12:00'
            )
        ]

        csv = CSVFile(convert_xls_to_csv(f, 'Sheet 2'), ['ID', 'Namä', 'Date'])

        assert list(csv.headers.keys()) == ['ID', 'Namä', 'Date']
        assert list(csv.lines) == [
            csv.rowtype(
                rownumber=2,
                id='1',
                nama='Döner',
                date='31.12.2015'
            ),
            csv.rowtype(
                rownumber=3,
                id='3',
                nama='“Cola”',
                date='31.12.2014 12:00'
            )
        ]


def test_empty_line_csv_file():
    data = (
        b'Datum, Reale Temperatur, Gef\xfchlte Temperatur\n'
        b'\n'
        b'02.01.2015, 0, kalt'
    )

    csv = CSVFile(
        BytesIO(data), ['datum', 'reale_temperatur', 'gefuhlte_temperatur']
    )

    csv.headers == ['datum', 'reale_temperatur', 'gefuhlte_temperatur']
    with pytest.raises(EmptyLineInFileError):
        list(csv.lines)

    # accept empty lines at the end
    data = (
        b'Datum, Reale Temperatur, Gef\xfchlte Temperatur\n'
        b'02.01.2015, 0, kalt'
        b'\n'
        b'\n'
    )

    csv = CSVFile(
        BytesIO(data), ['datum', 'reale_temperatur', 'gefuhlte_temperatur']
    )

    csv.headers == ['datum', 'reale_temperatur', 'gefuhlte_temperatur']
    assert list(csv.lines)


def test_match_headers_duplicate():
    with pytest.raises(DuplicateColumnNamesError):
        match_headers(['first_name', 'first_name'], expected=None)


def test_match_headers_order():
    matches = match_headers(
        headers=['firtst name', 'lastname'],
        expected=('first_name', 'last_name')
    )
    assert matches == ['first_name', 'last_name']

    matches = match_headers(
        headers=['firtst name', 'lastname'],
        expected=('last_name', 'first_name')
    )
    assert matches == ['first_name', 'last_name']


def test_match_headers_case():
    assert match_headers(['a', 'b'], expected=('A', 'B')) == ['A', 'B']
    assert match_headers(['a', 'b'], expected=('b', 'a')) == ['a', 'b']


def test_match_headers_missing():
    with pytest.raises(MissingColumnsError) as e:
        match_headers(['a', 'b'], expected=('a', 'c'))
    assert e.value.columns == ['c']

    with pytest.raises(MissingColumnsError) as e:
        match_headers(['ab', 'ba'], expected=('a', 'b'))
    assert e.value.columns == ['a', 'b']

    with pytest.raises(MissingColumnsError) as e:
        match_headers(['first', 'second'], expected=('first', 'third'))
    assert e.value.columns == ['third']

    assert match_headers(['a1', 'b2'], expected=('a1', 'c2')) == ['a1', 'c2']

    assert match_headers(['first', 'second'], expected=('first', 'sekond')) \
        == ['first', 'sekond']

    assert match_headers(['a', 'b', 'c'], expected=('a', 'c')) \
        == ['a', 'b', 'c']


def test_match_headers_ambiguous():
    with pytest.raises(AmbiguousColumnsError) as e:
        match_headers(['abcd', 'bcde'], expected=('bcd', ))

    assert list(e.value.columns.keys()) == ['bcd']
    assert set(e.value.columns['bcd']) == {'abcd', 'bcde'}


def test_convert_list_of_dicts_to_csv():
    data = [
        {
            'first_name': 'Dick',
            'last_name': 'Cheney'
        },
        {
            'first_name': 'Donald',
            'last_name': 'Rumsfeld'
        }
    ]

    # without providing a list of fields, the order of fields is random
    csv = convert_list_of_dicts_to_csv(data)
    header, dick, donald = csv.splitlines()

    assert 'first_name' in header
    assert 'last_name' in header

    assert 'Dick' in dick
    assert 'Cheney' in dick

    assert 'Donald' in donald
    assert 'Rumsfeld' in donald

    # we can change this by being explicit
    csv = convert_list_of_dicts_to_csv(data, ('first_name', 'last_name'))
    header, dick, donald = csv.splitlines()

    assert header == 'first_name,last_name'
    assert dick == 'Dick,Cheney'
    assert donald == 'Donald,Rumsfeld'

    csv = convert_list_of_dicts_to_csv(data, ('last_name', 'first_name'))
    header, dick, donald = csv.splitlines()

    assert header == 'last_name,first_name'
    assert dick == 'Cheney,Dick'
    assert donald == 'Rumsfeld,Donald'

    # or by providing a key
    csv = convert_list_of_dicts_to_csv(data, key=lambda f: f, reverse=True)
    header, dick, donald = csv.splitlines()

    assert header == 'last_name,first_name'
    assert dick == 'Cheney,Dick'
    assert donald == 'Rumsfeld,Donald'

    csv = convert_list_of_dicts_to_csv(data, key=lambda f: f, reverse=False)
    header, dick, donald = csv.splitlines()

    assert header == 'first_name,last_name'
    assert dick == 'Dick,Cheney'
    assert donald == 'Donald,Rumsfeld'


def test_convert_list_of_dicts_to_csv_escaping():
    data = [
        {
            'value': ',;"'
        }
    ]

    # without providing a list of fields, the order of fields is random
    csv = convert_list_of_dicts_to_csv(data)
    header, row = csv.splitlines()

    assert header == 'value'
    assert row == '",;"""'


def test_convert_list_of_dicts_to_xlsx():
    data = [
        {
            'first_name': 'Dick',
            'last_name': 'Cheney'
        },
        {
            'first_name': 'Donald',
            'last_name': 'Rumsfeld'
        }
    ]

    xlsx = convert_list_of_dicts_to_xlsx(data, ('first_name', 'last_name'))

    with tempfile.NamedTemporaryFile() as f:
        f.write(xlsx)

        rows = tuple(load_workbook(f).active.rows)

        assert rows[0][0].value == 'first_name'
        assert rows[0][1].value == 'last_name'
        assert rows[1][0].value == 'Dick'
        assert rows[1][1].value == 'Cheney'
        assert rows[2][0].value == 'Donald'
        assert rows[2][1].value == 'Rumsfeld'


def test_convert_irregular_list_of_dicts_to_csv():
    data = [
        {
            'name': 'Batman',
            'role': 'Superhero',
            'identity': 'Bruce Wayne'
        },
        {
            'name': 'Joker',
            'role': 'Supervillain',
            'gang': 'Injustice League'
        }
    ]

    # fields in random order
    csv = convert_list_of_dicts_to_csv(data)
    header, batman, joker = csv.splitlines()

    for column in ('name', 'role', 'identity', 'gang'):
        assert column in header

    for value in ('Batman', 'Superhero', 'Bruce Wayne'):
        assert value in batman

    for value in ('Joker', 'Supervillain', 'Injustice League'):
        assert value in joker

    # fields in defined order
    csv = convert_list_of_dicts_to_csv(
        data, ('name', 'role', 'identity', 'gang'))

    header, batman, joker = csv.splitlines()
    assert header == 'name,role,identity,gang'
    assert batman == 'Batman,Superhero,Bruce Wayne,'
    assert joker == 'Joker,Supervillain,,Injustice League'



import onegov.core
import os.path

from onegov.core import utils
from onegov.core.orm import SessionManager
from onegov.core.orm.types import HSTORE
from sqlalchemy import Column, Integer
from sqlalchemy.ext.declarative import declarative_base
from uuid import uuid4


def test_normalize_for_url():
    assert utils.normalize_for_url('asdf') == 'asdf'
    assert utils.normalize_for_url('Asdf') == 'asdf'
    assert utils.normalize_for_url('A S d f') == 'a-s-d-f'
    assert utils.normalize_for_url('far  away') == 'far-away'
    assert utils.normalize_for_url('währung') == 'waehrung'
    assert utils.normalize_for_url('grün') == 'gruen'
    assert utils.normalize_for_url('rötlich') == 'roetlich'
    assert utils.normalize_for_url('one/two') == 'one-two'
    assert utils.normalize_for_url('far / away') == 'far-away'
    assert utils.normalize_for_url('far <away>') == 'far-away'
    assert utils.normalize_for_url('far (away)') == 'far-away'
    assert utils.normalize_for_url('--ok--') == 'ok'
    assert utils.normalize_for_url('a...b..c.d') == 'a-b-c-d'


def test_lchop():
    assert utils.lchop('foobar', 'foo') == 'bar'
    assert utils.lchop('foobar', 'bar') == 'foobar'


def test_rchop():
    assert utils.rchop('foobar', 'foo') == 'foobar'
    assert utils.rchop('foobar', 'bar') == 'foo'
    assert utils.rchop('https://www.example.org/ex/amp/le', '/ex/amp/le') \
        == 'https://www.example.org'


def test_touch(temporary_directory):
    path = os.path.join(temporary_directory, 'test.txt')

    assert not os.path.isfile(path)

    utils.touch(path)

    assert os.path.isfile(path)

    with open(path, 'w') as f:
        f.write('asdf')

    utils.touch(path)

    with open(path, 'r') as f:
        assert f.read() == 'asdf'


def test_module_path():
    path = utils.module_path('onegov.core', 'utils.py')
    assert path == utils.module_path(onegov.core, 'utils.py')
    assert os.path.isfile(path)


def test_linkify():
    # this is really bleach's job, but we want to run the codepath anyway
    assert utils.linkify('info@example.org')\
        == '<a href="mailto:info@example.org">info@example.org</a>'
    assert utils.linkify('https://google.ch')\
        == '<a href="https://google.ch" rel="nofollow">https://google.ch</a>'

    # by default, linkify sanitizes the text before linkifying it
    assert utils.linkify('info@example.org<br>')\
        == '<a href="mailto:info@example.org">info@example.org</a>&lt;br&gt;'

    # we can disable that however
    assert utils.linkify('info@example.org<br>', escape=False)\
        == '<a href="mailto:info@example.org">info@example.org</a><br>'


def test_sanitize_html():
    # this is really bleach's job, but we want to run the codepath anyway
    assert utils.sanitize_html('<script>') == '&lt;script&gt;'


def test_increment_name():
    assert utils.increment_name('test') == 'test-1'
    assert utils.increment_name('test-2') == 'test-3'
    assert utils.increment_name('test2') == 'test2-1'
    assert utils.increment_name('test-1-1') == 'test-1-2'


def test_ensure_scheme():
    assert utils.ensure_scheme(None) is None
    assert utils.ensure_scheme('seantis.ch') == 'http://seantis.ch'
    assert utils.ensure_scheme('seantis.ch', 'https') == 'https://seantis.ch'

    assert utils.ensure_scheme('google.ch?q=onegov.cloud')\
        == 'http://google.ch?q=onegov.cloud'

    assert utils.ensure_scheme('https://abc.xyz') == 'https://abc.xyz'


def test_is_uuid():
    assert not utils.is_uuid(None)
    assert not utils.is_uuid('')
    assert not utils.is_uuid('asdf')
    assert utils.is_uuid(uuid4())
    assert utils.is_uuid(str(uuid4()))
    assert utils.is_uuid(uuid4().hex)


def test_is_non_string_iterable():
    assert utils.is_non_string_iterable([])
    assert utils.is_non_string_iterable(tuple())
    assert utils.is_non_string_iterable({})
    assert not utils.is_non_string_iterable('abc')
    assert not utils.is_non_string_iterable(b'abc')
    assert not utils.is_non_string_iterable(None)


def test_relative_url():
    assert utils.relative_url('https://www.google.ch/test') == '/test'
    assert utils.relative_url('https://usr:pwd@localhost:443/test') == '/test'
    assert utils.relative_url('/test') == '/test'
    assert utils.relative_url('/test?x=1&y=2') == '/test?x=1&y=2'
    assert utils.relative_url('/test?x=1&y=2#link') == '/test?x=1&y=2#link'


def test_is_subpath():
    assert utils.is_subpath('/', '/test')
    assert utils.is_subpath('/asdf', '/asdf/asdf')
    assert not utils.is_subpath('/asdf/', '/asdf')
    assert not utils.is_subpath('/a', '/b')
    assert not utils.is_subpath('/a', '/a/../b')


def test_is_sorted():
    assert utils.is_sorted('abc')
    assert not utils.is_sorted('aBc')
    assert utils.is_sorted('aBc', key=lambda i: i.lower())
    assert not utils.is_sorted('321')
    assert utils.is_sorted('321', reverse=True)


def test_get_unique_hstore_keys(postgres_dsn):

    Base = declarative_base()

    class Document(Base):
        __tablename__ = 'documents'

        id = Column(Integer, primary_key=True)
        _tags = Column(HSTORE, nullable=True)

        @property
        def tags(self):
            return set(self._tags.keys()) if self._tags else set()

        @tags.setter
        def tags(self, value):
            self._tags = {k: '' for k in value} if value else None

    mgr = SessionManager(postgres_dsn, Base)
    mgr.set_current_schema('foo')

    mgr.session().add(Document(tags=['foo', 'bar']))
    mgr.session().add(Document(tags=['foo', 'baz']))

    assert utils.get_unique_hstore_keys(mgr.session(), Document._tags) == {
        'foo', 'bar', 'baz'
    }



from onegov.core.orm.types import JSON
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import deferred
from sqlalchemy.schema import Column


class ContentMixin(object):
    """ Mixin providing a meta/content JSON pair. Meta is a JSON column loaded
    with each request, content is a JSON column loaded deferred (to be shown
    only in the detail view).

    """

    #: metadata associated with the form, for storing small amounts of data
    @declared_attr
    def meta(cls):
        return Column(JSON, nullable=False, default=dict)

    #: content associated with the form, for storing things like long texts
    @declared_attr
    def content(cls):
        return deferred(Column(JSON, nullable=False, default=dict))


def dictionary_based_property_factory(attribute):

    def dictionary_based_property(name):
        """ Enables access of dictionaries through properties.

        Usage::

            meta_property = dictionary_based_property_factory('meta')

            class Model(ContentMixin):
                content = meta_property('content')

        This will add a content getter, setter and deleter bound, which all
        proxy to meta['content'].

        Since this is just a thin wrapper around python's property we can also
        override the setter::

            class Model(ContentMixin):
                content = meta_property('content')

                @content.setter
                def set_content(self, value):
                    self.meta['content'] = value
                    self.meta['content_html'] = to_html(value)

                @content.deleter
                def del_content(self):
                    del self.meta['content']
                    del self.meta['content_html']

        meta_property and content_property, to access the
        :class:`ContentMixin`'s dictionaries are defined in this file.

        """

        def getter(self):
            dictionary = getattr(self, attribute, None)
            return dictionary.get(name) if dictionary is not None else None

        def setter(self, value):
            if getattr(self, attribute) is None:
                setattr(self, attribute, {})

            getattr(self, attribute)[name] = value

        def deleter(self):
            del getattr(self, attribute)[name]

        return property(getter, setter, deleter)

    return dictionary_based_property


meta_property = dictionary_based_property_factory('meta')
content_property = dictionary_based_property_factory('content')



import bleach
import hashlib
import inspect
import json
import magic
import mimetypes
import morepath
import os.path
import pydoc
import re

from collections import Iterable
from contextlib import contextmanager
from cProfile import Profile
from datetime import datetime
from functools import partial
from importlib import import_module
from itertools import groupby, tee
from purl import URL
from sqlalchemy import distinct
from unidecode import unidecode
from uuid import UUID
from webob import static


# http://stackoverflow.com/a/13500078
_unwanted_url_chars = re.compile(r'[\.\(\)\\/\s<>\[\]{},:;?!@&=+$#@%|\*"\'`]+')
_double_dash = re.compile(r'[-]+')
_number_suffix = re.compile(r'-([0-9]+)$')
_uuid = re.compile(
    r'[a-f0-9]{8}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[a-f0-9]{4}-?[a-f0-9]{12}')

# only temporary until bleach has a release > 1.4.1 -
_email_regex = re.compile((
    "([a-z0-9!#$%&'*+\/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+\/=?^_`"
    "{|}~-]+)*(@|\sat\s)(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?(\.|"
    "\sdot\s))+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?)"
))


def normalize_for_url(text):
    """ Takes the given text and makes it fit to be used for an url.

    That means replacing spaces and other unwanted characters with '-',
    lowercasing everything and turning unicode characters into their closest
    ascii equivalent using Unidecode.

    See https://pypi.python.org/pypi/Unidecode

    """

    # German is our main language, so we are extra considerate about it
    # (unidecode turns ü into u)
    text = text.replace("ü", "ue")
    text = text.replace("ä", "ae")
    text = text.replace("ö", "oe")
    clean = _unwanted_url_chars.sub('-', unidecode(text).strip(' ').lower())
    clean = _double_dash.sub('-', clean)
    clean = clean.strip('-')

    return clean


def increment_name(name):
    """ Takes the given name and adds a numbered suffix beginning at 1.

    For example::

        foo => foo-1
        foo-1 => foo-2

    """

    match = _number_suffix.search(name)
    number = (match and int(match.group(1)) or 0) + 1

    if match:
        return _number_suffix.sub('-{}'.format(number), name)
    else:
        return name + '-{}'.format(number)


def lchop(text, beginning):
    """ Removes the beginning from the text if the text starts with it. """

    if text.startswith(beginning):
        return text[len(beginning):]

    return text


def rchop(text, end):
    """ Removes the end from the text if the text ends with it. """

    if text.endswith(end):
        return text[:-len(end)]

    return text


@contextmanager
def profile(filename):
    """ Profiles the wrapped code and stores the result in the profiles folder
    with the given filename.

    """
    profiler = Profile()
    profiler.enable()

    yield

    profiler.disable()
    profiler.create_stats()
    profiler.dump_stats('profiles/{}'.format(filename))


@contextmanager
def timing(name=None):
    """ Runs the wrapped code and prints the time in ms it took to run it.
    The name is printed in front of the time, if given.

    """

    start = datetime.utcnow()

    yield

    duration = datetime.utcnow() - start
    duration = int(round(duration.total_seconds() * 1000, 0))

    if name:
        print('{}: {} ms'.format(name, duration))
    else:
        print('{} ms'.format(duration))


def module_path(module, subpath):
    """ Returns a subdirectory in the given python module.

    :module:
        A python module (actual module or string)

    :subpath:
        Subpath below that python module. Leading slashes ('/') are ignored.
    """

    subpath = subpath.strip('/')

    if isinstance(module, str):
        module = pydoc.locate(module)

    assert module is not None

    parent = os.path.dirname(inspect.getfile(module))
    path = os.path.join(parent, subpath)

    # always be paranoid with path manipulation
    assert is_subpath(parent, path)

    return path


def touch(file_path):
    """ Touches the file on the given path. """
    try:
        os.utime(file_path, None)
    except:
        open(file_path, 'a').close()


class Bunch(object):
    """ A simple but handy "collector of a bunch of named stuff" class.

    See `<http://code.activestate.com/recipes/\
    52308-the-simple-but-handy-collector-of-a-bunch-of-named/>`_.

    For example::

        point = Bunch(x=1, y=2)
        assert point.x == 1
        assert point.y == 2

        point.z = 3
        assert point.z == 3

    """
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


def render_file(file_path, request):
    """ Takes the given file_path (content) and renders it to the browser.
    The file must exist on the local system and be readable by the current
    process.

    """

    # this is a very cachable result - though it's possible that a file
    # changes it's content type, it should usually not, especially since
    # we emphasize the use of random filenames
    @request.app.cache.cache_on_arguments()
    def get_content_type(file_path):
        content_type = mimetypes.guess_type(file_path)[0]

        if not content_type:
            content_type = magic.from_file(file_path, mime=True)

        return content_type

    return request.get_response(
        static.FileApp(file_path, content_type=get_content_type(file_path)))


def hash_dictionary(dictionary):
    """ Computes a sha256 hash for the given dictionary. The dictionary
    is expected to only contain values that can be serialized by json.

    That includes int, decimal, string, boolean.

    Note that this function is not meant to be used for hashing secrets. Do
    not include data in this dictionary that is secret!

    """
    dict_as_string = json.dumps(dictionary, sort_keys=True).encode('utf-8')
    return hashlib.sha1(dict_as_string).hexdigest()


def groupbylist(*args, **kwargs):
    """ Works just like Python's ``itertools.groupby`` function, but instead
    of returning generators, it returns lists.

    """
    return [(k, list(g)) for k, g in groupby(*args, **kwargs)]


def sanitize_html(html):
    """ Takes the given html and strips all but a whitelisted number of tags
    from it.

    """

    if not html:
        return html

    allowed_tags = [
        'a',
        'abbr',
        'b',
        'br',
        'blockquote',
        'code',
        'del',
        'div',
        'em',
        'i',
        'img',
        'hr',
        'li',
        'ol',
        'p',
        'strong',
        'sup',
        'ul',
        'h1',
        'h2',
        'h3',
        'h4',
        'h5',
        'h6'
    ]

    allowed_attributes = {
        'a': ['href', 'title'],
        'abbr': ['title'],
        'acronym': ['title'],
        'img': ['src', 'alt', 'title']
    }

    return bleach.clean(html, tags=allowed_tags, attributes=allowed_attributes)


def linkify(text, escape=True):
    """ Takes plain text and injects html links for urls and email addresses.

    By default the text is html escaped before it is linkified. This accounts
    for the fact that we usually use this for text blocks that we mean to
    extend with email addresses and urls.

    If html is already possible, why linkify it?

    Note: We need to clean the html after we've created it (linkify
    parses escaped html and turns it into real html). As a consequence it
    is possible to have html urls in the text that won't be escaped.

    """

    if not text:
        return text

    # do not parse email until this is fixed:
    # https://github.com/jsocol/bleach/issues/154

    # .. use a simple substitute instead
    text = emailify(text)

    linkified = bleach.linkify(text, parse_email=False)

    if not escape:
        return linkified

    return bleach.clean(
        linkified, tags=['a'], attributes={'a': ['href', 'rel']})


def emailify(text):
    # the link generating code below is not necessarily safe, so we need
    # to clean the result of each generation
    clean = partial(bleach.clean, tags=['a'], attributes={'a': ['href']})

    emails = (
        email[0] for email in re.findall(_email_regex, text)
        if not email[0].startswith('//')
    )

    for email in emails:
        link = '<a href="mailto:{email}">{email}</a>'.format(email=email)
        link = clean(link)

        text = text.replace(email, link)

    return text


def ensure_scheme(url, default='http'):
    """ Makes sure that the given url has a scheme in front, if none
    was provided.

    """

    if not url:
        return url

    # purl (or to be precise urlparse) will parse empty host names ('abc.xyz')
    # wrongly, assuming the abc.xyz is a path. by adding a double slash if
    # there isn't one already, we can circumvent that problem
    if '//' not in url:
        url = '//' + url

    _url = URL(url)

    if _url.scheme():
        return url

    return _url.scheme(default).as_string()


def is_uuid(value):
    """ Returns true if the given value is a uuid. The value may be a string
    or of type UUID. If it's a string, the uuid is checked with a regex.
    """
    if isinstance(value, str):
        return _uuid.match(str(value))

    return isinstance(value, UUID)


def is_non_string_iterable(obj):
    """ Returns true if the given obj is an iterable, but not a string. """
    return not (isinstance(obj, str) or isinstance(obj, bytes))\
        and isinstance(obj, Iterable)


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def relative_url(absolute_url):
    """ Removes everything in front of the path, including scheme, host,
    username, password and port.

    """
    url = URL._mutate(
        URL(absolute_url),
        scheme=None,
        username=None,
        password=None,
        host=None,
        port=None
    )

    return url.as_string()


def is_subpath(directory, path):
    """ Returns true if the given path is inside the given directory. """
    directory = os.path.join(os.path.realpath(directory), '')
    path = os.path.realpath(path)

    # return true, if the common prefix of both is equal to directory
    # e.g. /a/b/c/d.rst and directory is /a/b, the common prefix is /a/b
    return os.path.commonprefix([path, directory]) == directory


def is_sorted(iterable, key=lambda i: i, reverse=False):
    """ Returns True if the iterable is sorted. """

    i1, i2 = tee(iterable)

    for a, b in zip(i1, sorted(i2, key=key, reverse=reverse)):
        if a is not b:
            return False

    return True


def morepath_modules(cls):
    """ Returns all morepath modules which should be scanned for the given
    morepath application class.

    We can't reliably know the actual morepath modules that
    need to be scanned, which is why we assume that each module has
    one namespace (like 'more.transaction' or 'onegov.core').

    """
    for base in cls.__mro__:
        if not issubclass(base, morepath.App):
            continue

        if base is morepath.App:
            continue

        module = '.'.join(base.__module__.split('.')[:2])

        if module.startswith('test'):
            continue

        yield module


def scan_morepath_modules(cls):
    """ Tries to scann all the morepath modules required for the given
    application class. This is not guaranteed to stay reliable as there is
    no sure way to discover all modules required by the application class.

    """
    for module in sorted(morepath_modules(cls)):
        morepath.scan(import_module(module))


def get_unique_hstore_keys(session, column):
    """ Returns a set of keys found in an hstore column over all records
    of its table.

    """

    return {
        key
        for row in session.query(distinct(column.keys())).all()
        for key in row[0]
    }



import pytest
import requests

from datetime import datetime
from freezegun import freeze_time
from morepath.error import ConflictError
from onegov.core import cronjobs, Framework
from onegov.core.utils import scan_morepath_modules
from pytest_localserver.http import WSGIServer
from sedate import ensure_timezone, replace_timezone
from sqlalchemy.ext.declarative import declarative_base
from time import sleep
from webtest import TestApp as Client


def test_is_scheduled_at():
    job = cronjobs.Job(lambda: None, hour=8, minute=0, timezone='CET')

    assert job.hour == 8
    assert job.minute == 0
    assert job.timezone == ensure_timezone('CET')
    assert job.name == 'test_is_scheduled_at.<locals>.<lambda>'

    # dates must be timezone aware
    assert job.is_scheduled_at(
        replace_timezone(datetime(2015, 1, 1, 8), 'CET'))

    # though we may also use this shortcut for the same thing
    assert job.is_scheduled_at(datetime(2015, 1, 1, 8), 'CET')

    # if there's no timezone we fail
    with pytest.raises(AssertionError):
        job.is_scheduled_at(datetime(2015, 1, 1, 8))

    assert not job.is_scheduled_at(datetime(2015, 1, 1, 7, 59), 'CET')
    assert job.is_scheduled_at(datetime(2015, 1, 1, 8, 0, 1), 'CET')
    assert job.is_scheduled_at(datetime(2015, 1, 1, 8, 0, 59), 'CET')
    assert not job.is_scheduled_at(datetime(2015, 1, 1, 8, 1), 'CET')

    assert not job.is_scheduled_at(datetime(2015, 1, 1, 8, 0, 1), 'UTC')
    assert not job.is_scheduled_at(datetime(2015, 1, 1, 8, 0, 59), 'UTC')
    assert job.is_scheduled_at(datetime(2015, 1, 1, 7, 0, 1), 'UTC')
    assert job.is_scheduled_at(datetime(2015, 1, 1, 7, 0, 59), 'UTC')


def test_overlapping_cronjobs():

    class App(Framework):
        pass

    @App.cronjob(hour=8, minute=0, timezone='UTC')
    def first_job(request):
        pass

    @App.cronjob(hour=8, minute=0, timezone='UTC')
    def second_job(request):
        pass

    scan_morepath_modules(App)

    with pytest.raises(ConflictError):
        App.commit()


def test_non_5_minutes_cronjobs():

    class App(Framework):
        pass

    @App.cronjob(hour=8, minute=1, timezone='UTC')
    def first_job(request):
        pass

    scan_morepath_modules(App)

    with pytest.raises(AssertionError):
        App.commit()


def test_cronjobs_integration(postgres_dsn):

    result = 0
    cronjobs.CRONJOB_POLL_RESOLUTION = 1

    class App(Framework):
        pass

    @App.path(path='')
    class Root(object):
        pass

    @App.json(model=Root)
    def view_root(self, request):
        return {}

    @App.cronjob(hour=8, minute=0, timezone='UTC')
    def run_test_cronjob(request):
        nonlocal result
        result += 1

    scan_morepath_modules(App)

    app = App()
    app.configure_application(dsn=postgres_dsn, base=declarative_base())
    app.namespace = 'municipalities'
    app.set_application_id('municipalities/new-york')

    # to test we need an actual webserver, webtest doesn't cut it here because
    # we are making requests from the application itself
    server = WSGIServer(application=app)

    try:
        server.start()

        with freeze_time(replace_timezone(datetime(2016, 1, 1, 8, 0), 'UTC')):
            requests.get(server.url)
            sleep(2.5)

    finally:
        server.stop()

    assert result == 1


def test_disable_cronjobs():

    class App(Framework):
        pass

    @App.path(path='')
    class Root(object):
        pass

    @App.json(model=Root)
    def view_root(self, request):
        return {}

    @App.cronjob(hour=8, minute=0, timezone='UTC')
    def run_test_cronjob(request):
        pass

    @App.setting(section='cronjobs', name='enabled')
    def cronjobs_enabled():
        return False

    scan_morepath_modules(App)

    app = App()
    app.configure_application()
    app.namespace = 'municipalities'
    app.set_application_id('municipalities/new-york')

    client = Client(app)
    client.get('/')

    assert not app.config.cronjob_registry.cronjob_threads



import importlib
import networkx as nx
import pkg_resources
import transaction

from alembic.migration import MigrationContext
from alembic.operations import Operations
from inspect import getmembers, isfunction, ismethod
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import JSON
from sqlalchemy import Column, Text
from sqlalchemy.engine.reflection import Inspector


class UpgradeState(Base, TimestampMixin):
    """ Keeps the state of all upgrade steps over all modules. """

    __tablename__ = 'upgrades'

    # the name of the module (e.g. onegov.core)
    module = Column(Text, primary_key=True)

    # a json holding the state of the upgrades
    state = Column(JSON, nullable=False, default=dict)

    @property
    def executed_tasks(self):
        if not self.state:
            return set()
        else:
            return set(self.state.get('executed_tasks', []))

    def was_already_executed(self, task):
        return task.task_name in self.executed_tasks

    def mark_as_executed(self, task):
        if not self.state:
            self.state = {}

        if 'executed_tasks' in self.state:
            self.state['executed_tasks'].append(task.task_name)
            self.state['executed_tasks'] = list(
                set(self.state['executed_tasks']))
        else:
            self.state['executed_tasks'] = [task.task_name]

        self.state.changed()


def get_distributions_with_entry_map(key):
    """ Iterates through all distributions with entry_maps and yields
    each distribution along side the entry map with the given key.

    """
    for distribution in pkg_resources.working_set:
        if hasattr(distribution, 'get_entry_map'):
            entry_map = distribution.get_entry_map(key)

            if entry_map:
                yield distribution, entry_map


def get_upgrade_modules():
    """ Returns all modules that registered themselves for onegov.core
    upgrades like this::

        entry_points={
            'onegov.core': [
                'upgrade = onegov.mypackage.upgrade'
            ]
        }

    """
    for distribution, entry_map in get_distributions_with_entry_map('onegov'):
        if 'upgrade' in entry_map:
            yield (
                '.'.join(entry_map['upgrade'].module_name.split('.')[:2]),
                importlib.import_module(entry_map['upgrade'].module_name)
            )


class upgrade_task(object):
    """ Marks the decorated function as an upgrade task. Upgrade tasks should
    be defined outside classes (except for testing) - that is in the root of
    the module (directly in onegov/form/upgrades.py for example).

    Each upgrade task has the following properties:

    :name:
        The name of the task in readable English. Do not change the name of
        the task once you pushed your changes. This name is used as an id!

    :always_run:
        By default, tasks are run once on if they haven't been run yet, or if
        the module was not yet under upgrade control (it just added itself
        through the entry_points mechanism).

        There are times where this is not wanted.

        If you enable upgrades in your module and you plan to run an upgrade
        task in the same release, then you need to 'always_run' and detect
        yourself if you need to do any changes.

        If you need to enable upgrades and run a task at the same time which
        can only be run once and is impossible to make idempotent, then you
        need to make two releases. One that enables upgade control and a second
        one that

        Also, if you want to run an upgrade task every time you upgrade (for
        example, for maintenance), then you should also use it.

        The easiest way to use upgrade control is to enable it from the first
        release on.

        Another way to tackle this is to write update methods idempotent as
        often as possible and to always run them.

    :requires:
        A reference to another projects upgrade task that should be run first.

        For example::

            requires='onegov.form:Add new form field'

        If the reference cannot be found, the upgrade fails.

    Always run tasks may return ``False`` if they wish to be considered
    as not run (and therefore not shown in the upgrade runner).

    """

    def __init__(self, name, always_run=False, requires=None):
        self.name = name
        self.always_run = always_run
        self.requires = requires

    def __call__(self, fn):
        fn.task_name = self.name
        fn.always_run = self.always_run
        fn.requires = self.requires
        return fn


def is_task(function):
    """ Returns True if the given function is an uprade task. """
    if not (isfunction(function) or ismethod(function)):
        return False

    return hasattr(function, 'task_name')


def get_module_tasks(module):
    """ Goes through a module or class and returns all upgrade tasks. """
    for name, function in getmembers(module, is_task):
        yield function


def get_tasks_by_id(upgrade_modules=None):
    """ Takes a list of upgrade modules or classes and returns the tasks
    keyed by id.

    """

    tasks = {}
    fn_names = set()

    for distribution, upgrade_module in upgrade_modules:

        for function in get_module_tasks(upgrade_module):
            task_id = ':'.join((distribution, function.task_name))

            assert task_id not in tasks, "Duplicate task"
            tasks[task_id] = function

            # make sure we don't have duplicate function names - it works, but
            # it makes debugging harder
            assert function.__name__ not in fn_names, "Duplicate function name"
            fn_names.add(function.__name__)

    return tasks


def get_tasks(upgrade_modules=None):
    """ Takes a list of upgrade modules or classes and returns the
    tasks that should be run in the order they should be run.

    The order takes dependencies into account.

    """

    tasks = get_tasks_by_id(upgrade_modules or get_upgrade_modules())
    ordered_tasks = []
    roots = set()
    graph = nx.DiGraph()

    for task_id, task in tasks.items():
        if task.requires is None:
            roots.add(task_id)
        else:
            graph.add_edge(task.requires, task_id)

    for task_id in sorted(roots):
        ordered_tasks.append((task_id, tasks[task_id]))

        try:
            for _, task_id in nx.dfs_edges(graph, task_id):
                ordered_tasks.append((task_id, tasks[task_id]))
        except KeyError:
            pass

    return ordered_tasks


def register_modules(session, modules, tasks):
    """ Sets up the state tracking for all modules. Initially, all tasks
    are marekd as executed, because we assume tasks to upgrade older
    deployments to a new deployment.

    If this is a new deployment we do not need to execute these tasks.

    Tasks where this is not desired should be marked as 'always_run'.
    They will then manage their own state (i.e. check if they need to
    run or not).

    This function is idempotent.

    """

    for module, upgrade_module in modules:
        query = session.query(UpgradeState)
        query = query.filter(UpgradeState.module == module)

        if query.first():
            continue

        session.add(
            UpgradeState(module=module, state={
                'executed_tasks': [
                    task.task_name for task_id, task in tasks
                    if task_id.startswith(module)
                ]
            })
        )

    session.flush()


def register_all_modules_and_tasks(session):
    """ Registers all the modules and all the tasks. """
    register_modules(session, get_upgrade_modules(), get_tasks())


class UpgradeTransaction(object):
    """ Holds the session and the alembic operations connection together and
    commits/aborts both at the same time.

    Not really a two-phase commit solution. That could be accomplished but
    for now this suffices.

    """

    def __init__(self, context):
        self.operations_transaction = context.operations_connection.begin()
        self.session = context.session

    def flush(self):
        self.session.flush()

    def commit(self):
        self.operations_transaction.commit()
        transaction.commit()

    def abort(self):
        self.operations_transaction.rollback()
        transaction.abort()


class UpgradeContext(object):
    """ Holdes the context of the upgrade. An instance of this is passed
    to each upgrade task.

    """

    def __init__(self, request):
        self.request = request
        self.session = request.app.session()
        self.app = request.app
        self.schema = request.app.session_manager.current_schema
        self.engine = self.session.bind

        # The locale of the upgrade is the default locale
        request.app.session_manager.set_locale(
            current_locale=self.request.locale,
            default_locale=self.request.default_locale
        )

        # Make sure the connection is the same for the session, the engine
        # and the alembic migrations context. Otherwise the upgrade locks up.
        # If that happens again, test with this:
        #
        #       http://docs.sqlalchemy.org/en/latest/core/pooling.html
        #       #sqlalchemy.pool.AssertionPool
        #
        self.operations_connection = self.session._connection_for_bind(
            self.session.bind)
        self.operations = Operations(
            MigrationContext.configure(self.operations_connection))

    def begin(self):
        return UpgradeTransaction(self)

    def has_column(self, table, column):
        inspector = Inspector(self.operations_connection)
        return column in {c['name'] for c in inspector.get_columns(table)}


class UpgradeRunner(object):
    """ Runs all tasks of a :class:`UpgradeTasksRegistry`. """

    def __init__(self, modules, tasks, commit=True,
                 on_task_success=None, on_task_fail=None):
        self.modules = modules
        self.tasks = tasks
        self.commit = commit
        self.states = {}

        self._on_task_success = on_task_success
        self._on_task_fail = on_task_fail

    def on_task_success(self, task):
        if self._on_task_success is not None:
            self._on_task_success(task)

    def on_task_fail(self, task):
        if self._on_task_fail is not None:
            self._on_task_fail(task)

    def get_state(self, context, module):
        if module not in self.states:
            query = context.session.query(UpgradeState)
            query = query.filter(UpgradeState.module == module)

            # assume that all modules have been ran through register_module
            self.states[module] = query.one()

        return context.session.merge(self.states[module], load=False)

    def register_modules(self, request):
        register_modules(request.app.session(), self.modules, self.tasks)

        if self.commit:
            transaction.commit()

    def get_module_from_task_id(self, task_id):
        return task_id.split(':')[0]

    def run_upgrade(self, request):
        self.register_modules(request)

        tasks = ((self.get_module_from_task_id(i), t) for i, t in self.tasks)
        executed = 0

        for module, task in tasks:
            context = UpgradeContext(request)
            state = self.get_state(context, module)

            if not task.always_run and state.was_already_executed(task):
                continue

            upgrade = context.begin()

            try:
                result = task(context)

                # mark all tasks as executed, even 'always run' ones
                state.mark_as_executed(task)
                upgrade.flush()

                # always-run tasks which return False are considered
                # to have not run
                hidden = task.always_run and result is False

                if not hidden:
                    executed += 1
                    self.on_task_success(task)

            except:
                upgrade.abort()
                self.on_task_fail(task)

                raise
            else:
                if self.commit:
                    upgrade.commit()
                else:
                    upgrade.abort()
            finally:
                context.session.invalidate()
                context.engine.dispose()

        return executed



import json
import morepath
import os
import transaction
import pytest

from base64 import b64decode
from datetime import datetime, timedelta
from email.header import decode_header
from email.utils import parseaddr
from freezegun import freeze_time
from onegov.core import Framework
from onegov.core.upgrade import UpgradeState
from onegov.core.mail import convert_to_plaintext
from onegov.server import Config, Server
from webtest import TestApp as Client
from wtforms import Form, StringField


def test_set_application_id():
    app = Framework()
    app.namespace = 'namespace'
    app.configure_application()
    app.set_application_id('namespace/id')

    assert app.schema == 'namespace-id'


def test_configure():
    class App(Framework):
        configured = []

        def configure_a(self, **cfg):
            self.configured.append('a')

        def configure_b(self, **cfg):
            self.configured.append('b')

        @property
        def configure_c(self, **cfg):
            self.configured.append('c')

    app = App()
    app.configure_application()

    assert app.configured == ['a', 'b']


def test_virtual_host_request():

    class App(Framework):
        pass

    @App.path(path='/')
    class Root(object):
        pass

    @App.path(path='/blog')
    class Blog(object):
        pass

    @App.view(model=Root)
    def view_root(self, request):
        return request.link(self) + ' - root'

    @App.view(model=Blog)
    def view_blog(self, request):
        return request.link(self) + ' - blog'

    app = App()
    app.configure_application()

    c = Client(app)

    response = c.get('/')
    assert response.body == b'http://localhost/ - root'

    response = c.get('/blog')
    assert response.body == b'http://localhost/blog - blog'

    # X_VHM_HOST is a simple prefix..
    response = c.get('/', headers={'X_VHM_HOST': 'http://example.org'})
    assert response.body == b'http://example.org/ - root'

    response = c.get('/blog', headers={'X_VHM_HOST': 'http://example.org'})
    assert response.body == b'http://example.org/blog - blog'

    # .. though it won't lead to '//'s in the url
    response = c.get('/', headers={'X_VHM_HOST': 'http://example.org/'})
    assert response.body == b'http://example.org/ - root'

    # X_VHM_ROOT set to '/' has no influence
    response = c.get('/', headers={'X_VHM_ROOT': '/'})
    assert response.body == b'http://localhost/ - root'

    # just like X_VHM_HOST it tries to not introduce any '//'s
    response = c.get('/blog', headers={'X_VHM_ROOT': '/blog'})
    assert response.body == b'http://localhost/ - blog'

    response = c.get('/blog', headers={'X_VHM_ROOT': '/blog/'})
    assert response.body == b'http://localhost/ - blog'

    # X_VHM_HOST and X_VHM_ROOT may be used together
    response = c.get('/blog', headers={
        'X_VHM_ROOT': '/blog', 'X_VHM_HOST': 'https://blog.example.org/'})
    assert response.body == b'https://blog.example.org/ - blog'

    response = c.get('/blog', headers={
        'X_VHM_ROOT': '/blog',
        'X_VHM_HOST': 'https://blog.example.org/'})
    assert response.body == b'https://blog.example.org/ - blog'


def test_registered_upgrade_tasks(postgres_dsn):

    app = Framework()

    app.configure_application(dsn=postgres_dsn)
    app.namespace = 'foo'
    app.set_application_id('foo/bar')

    assert app.session().query(UpgradeState).count() > 0


def test_browser_session_request():

    class App(Framework):
        pass

    @App.path(path='/')
    class Root(object):
        pass

    @App.view(model=Root)
    def view_root(self, request):
        # the session id is only available if there's a change in the
        # browser session
        request.browser_session['foo'] = 'bar'
        return request.cookies['session_id']

    @App.view(model=Root, name='login')
    def view_login(self, request):
        request.browser_session.logged_in = True
        return 'logged in'

    @App.view(model=Root, name='status')
    def view_status(self, request):
        if request.browser_session.has('logged_in'):
            return 'logged in'
        else:
            return 'logged out'

    app = App()
    app.application_id = 'test'
    app.configure_application(identity_secure=False)  # allow http

    app.cache_backend = 'dogpile.cache.memory'
    app.cache_backend_arguments = {}

    c1 = Client(app)
    c2 = Client(app)

    c1.get('/').text == c1.get('/').text
    c1.get('/').text != c2.get('/').text
    c2.get('/').text == c2.get('/').text

    c1.get('/status').text == 'logged out'
    c2.get('/status').text == 'logged out'

    c1.get('/login')

    c1.get('/status').text == 'logged in'
    c2.get('/status').text == 'logged out'

    app.application_id = 'tset'
    c1.get('/status').text == 'logged out'

    app.application_id = 'test'
    c1.get('/status').text == 'logged in'


def test_browser_session_dirty():

    class App(Framework):
        pass

    @App.path(path='/')
    class Root(object):
        pass

    @App.view(model=Root, name='undirty')
    def view_undirty(self, request):
        request.browser_session.get('foo')
        return ''

    @App.view(model=Root, name='dirty')
    def view_dirty(self, request):
        request.browser_session['foo'] = 'bar'
        return ''

    app = App()
    app.application_id = 'test'
    app.configure_application(identity_secure=False)  # allow http

    app.cache_backend = 'dogpile.cache.memory'
    app.cache_backend_arguments = {}

    client = Client(app)

    client.get('/undirty')
    assert 'session_id' not in client.cookies

    client.get('/dirty')
    assert 'session_id' in client.cookies

    old_session_id = client.cookies['session_id']

    client.get('/undirty')
    assert client.cookies['session_id'] == old_session_id

    client.get('/undirty')
    assert client.cookies['session_id'] == old_session_id


def test_request_messages():

    class App(Framework):
        pass

    @App.path(path='/')
    class Root(object):
        pass

    class Message(object):
        def __init__(self, text, type):
            self.text = text
            self.type = type

    @App.path(model=Message, path='/messages')
    def get_message(text, type):
        return Message(text, type)

    @App.view(model=Message, name='add')
    def view_add_message(self, request):
        request.message(self.text, self.type)

    @App.view(model=Root)
    def view_root(self, request):
        return json.dumps(list(request.consume_messages()))

    app = App()
    app.application_id = 'test'
    app.configure_application(identity_secure=False)  # allow http

    app.cache_backend = 'dogpile.cache.memory'
    app.cache_backend_arguments = {}

    c1 = Client(app)
    c2 = Client(app)
    c1.get('/messages/add?text=one&type=info')
    c1.get('/messages/add?text=two&type=warning')
    c2.get('/messages/add?text=three&type=error')

    messages = json.loads(c1.get('/').text)
    assert len(messages) == 2
    assert messages[0][0] == 'one'
    assert messages[1][0] == 'two'
    assert messages[0][1] == 'info'
    assert messages[1][1] == 'warning'

    messages = json.loads(c1.get('/').text)
    assert len(messages) == 0

    messages = json.loads(c2.get('/').text)
    assert len(messages) == 1

    assert messages[0][0] == 'three'
    assert messages[0][1] == 'error'

    messages = json.loads(c2.get('/').text)
    assert len(messages) == 0


def test_fix_webassets_url():

    import onegov.core
    import more.transaction
    import more.webassets
    morepath.scan(more.transaction)
    morepath.scan(more.webassets)
    morepath.scan(onegov.core)

    class App(Framework):
        pass

    @App.path(path='/')
    class Root(object):
        pass

    @App.html(model=Root)
    def view_root(self, request):
        return '/' + request.app.config.webasset_registry.url + '/jquery.js'

    class TestServer(Server):

        def configure_morepath(self, *args, **kwargs):
            pass

    server = TestServer(Config({
        'applications': [
            {
                'path': '/towns/*',
                'application': App,
                'namespace': 'towns'
            }
        ]
    }))

    client = Client(server)

    # more.webassets doesn't know about virtual hosting (that is to say
    # Morepath does not know about it).
    #
    # Since it wants to create urls for the root of the application ('/'),
    # it will create something like '/xxx/jquery.js'
    #
    # We really want '/towns/test/xxx' here, which is something the onegov
    # core Framework application fixes through a tween.
    response = client.get('/towns/test')
    assert response.body == (b'http://localhost/towns/test/'
                             b'7da9c72a3b5f9e060b898ef7cd714b8a/jquery.js')


def test_sign_unsign():
    framework = Framework()
    framework.identity_secret = 'test'
    framework.application_id = 'one'

    assert framework.sign('foo').startswith('foo.')
    assert framework.unsign(framework.sign('foo')) == 'foo'

    signed_by_one = framework.sign('foo')
    framework.application_id = 'two'
    assert framework.unsign(signed_by_one) is None
    framework.application_id = 'one'
    assert framework.unsign(signed_by_one) == 'foo'

    signed = framework.sign('foo')
    framework.identity_secret = 'asdf'
    assert framework.unsign(signed) is None

    signed = framework.sign('foo')
    framework.unsign('bar' + signed) is None


def test_csrf_secret_key():
    app = Framework()

    with pytest.raises(AssertionError):
        app.configure_application(identity_secret='test', csrf_secret='test')

    with pytest.raises(AssertionError):
        app.configure_application(identity_secret='very-secret-key')

    with pytest.raises(AssertionError):
        app.configure_application(csrf_secret='another-very-secret-key')

    app.configure_application(identity_secret='x', csrf_secret='y')


def test_csrf():

    class MyForm(Form):
        name = StringField('Name')

    class App(Framework):
        pass

    @App.path(path='/')
    class Root(object):
        pass

    @App.view(model=Root, request_method='GET')
    def view_get_root(self, request):
        form = request.get_form(MyForm, i18n_support=False)
        return form.csrf_token._value()

    @App.view(model=Root, request_method='POST')
    def view_post_root(self, request):
        if request.get_form(MyForm, i18n_support=False).validate():
            return 'success'
        else:
            return 'fail'

    app = App()
    app.application_id = 'test'
    app.configure_application(
        identity_secure=False,
        disable_memcached=True,
        csrf_time_limit=60
    )

    client = Client(app)
    csrf_token = client.get('/').text
    assert client.post('/', {'csrf_token': csrf_token}).text == 'success'
    assert client.post('/', {'csrf_token': csrf_token + 'x'}).text == 'fail'
    assert client.post('/', {'csrf_token': csrf_token}).text == 'success'

    with freeze_time(datetime.now() + timedelta(minutes=2)):
        assert client.post('/', {'csrf_token': csrf_token}).text == 'fail'


def test_send_email(smtp):

    app = Framework()
    app.mail_host, app.mail_port = smtp.address
    app.mail_sender = 'noreply@example.org'
    app.mail_force_tls = False
    app.mail_username = None
    app.mail_password = None
    app.mail_use_directory = False

    app.send_email(
        reply_to='info@example.org',
        receivers=['recipient@example.org'],
        subject="Test E-Mail",
        content="This e-mail is just a test"
    )

    transaction.commit()

    assert len(smtp.outbox) == 1
    message = smtp.outbox[0]

    assert message['Sender'] == 'noreply@example.org'
    assert message['From'] == 'noreply@example.org'
    assert message['Reply-To'] == 'info@example.org'
    assert message['Subject'] == 'Test E-Mail'
    assert message.get_payload()[0].as_string() == (
        'Content-Type: text/plain; charset="utf-8"\n'
        'MIME-Version: 1.0\n'
        'Content-Transfer-Encoding: base64\n\n'
        'VGhpcyBlLW1haWwgaXMganVzdCBhIHRlc3Q=\n'
    )
    assert message.get_payload()[1].as_string() == (
        'Content-Type: text/html; charset="utf-8"\n'
        'MIME-Version: 1.0\n'
        'Content-Transfer-Encoding: base64\n\n'
        'VGhpcyBlLW1haWwgaXMganVzdCBhIHRlc3Q=\n'
    )


def test_send_email_with_name(smtp):

    app = Framework()
    app.mail_host, app.mail_port = smtp.address
    app.mail_sender = 'noreply@example.org'
    app.mail_force_tls = False
    app.mail_username = None
    app.mail_password = None
    app.mail_use_directory = False

    app.send_email(
        reply_to='Govikon <info@example.org>',
        receivers=['recipient@example.org'],
        subject="Test E-Mail",
        content="This e-mail is just a test"
    )

    transaction.commit()

    assert len(smtp.outbox) == 1
    message = smtp.outbox[0]

    assert message['Sender'] == 'Govikon <noreply@example.org>'
    assert message['From'] == 'Govikon <noreply@example.org>'
    assert message['Reply-To'] == 'Govikon <info@example.org>'
    assert message['Subject'] == 'Test E-Mail'
    assert message.get_payload()[0].as_string() == (
        'Content-Type: text/plain; charset="utf-8"\n'
        'MIME-Version: 1.0\n'
        'Content-Transfer-Encoding: base64\n\n'
        'VGhpcyBlLW1haWwgaXMganVzdCBhIHRlc3Q=\n'
    )
    assert message.get_payload()[1].as_string() == (
        'Content-Type: text/html; charset="utf-8"\n'
        'MIME-Version: 1.0\n'
        'Content-Transfer-Encoding: base64\n\n'
        'VGhpcyBlLW1haWwgaXMganVzdCBhIHRlc3Q=\n'
    )


def test_send_email_to_maildir(temporary_directory):

    app = Framework()
    app.mail_sender = 'noreply@example.org'
    app.mail_use_directory = True
    app.mail_directory = temporary_directory

    app.send_email(
        reply_to='Govikon <info@example.org>',
        receivers=['recipient@example.org'],
        subject="Test E-Mail",
        content="This e-mail is just a test"
    )

    assert not os.listdir(temporary_directory)

    transaction.commit()

    assert set(os.listdir(temporary_directory)) == {'cur', 'new', 'tmp'}

    new_emails = os.path.join(temporary_directory, 'new')
    assert len(os.listdir(new_emails)) == 1

    new_email = os.path.join(new_emails, os.listdir(new_emails)[0])
    with open(new_email, 'r') as f:
        email = f.read()

    assert 'Subject: Test E-Mail' in email
    assert 'Reply-To: Govikon <info@example.org>' in email
    assert 'From: Govikon <noreply@example.org>' in email
    assert 'Sender: Govikon <noreply@example.org>' in email
    assert 'To: recipient@example.org' in email


def test_send_email_iso_8859_1(smtp):
    app = Framework()
    app.mail_host, app.mail_port = smtp.address
    app.mail_sender = 'noreply@example.org'
    app.mail_force_tls = False
    app.mail_username = None
    app.mail_password = None
    app.mail_use_directory = False

    app.send_email(
        reply_to='Gövikon <info@example.org>',
        receivers=['recipient@example.org'],
        subject="Nüws",
        content="This e-mäil is just a test"
    )

    assert len(smtp.outbox) == 0
    transaction.commit()

    assert len(smtp.outbox) == 1
    message = smtp.outbox[0]

    def decode(header):
        name, addr = parseaddr(header)

        try:
            name = decode_header(name)[0][0].decode('utf-8')
        except AttributeError:
            pass

        return name, addr

    assert decode(message['Sender']) == ("Gövikon", 'noreply@example.org')
    assert decode(message['From']) == ("Gövikon", 'noreply@example.org')
    assert decode(message['Reply-To']) == ("Gövikon", 'info@example.org')

    assert decode_header(message['Subject'])[0][0].decode('utf-8') \
        == "Nüws"

    assert message.get_payload()[0].as_string() == (
        'Content-Type: text/plain; charset="utf-8"\n'
        'MIME-Version: 1.0\n'
        'Content-Transfer-Encoding: base64\n\n'
        'VGhpcyBlLW3DpGlsIGlzIGp1c3QgYSB0ZXN0\n'
    )
    assert message.get_payload()[1].as_string() == (
        'Content-Type: text/html; charset="utf-8"\n'
        'MIME-Version: 1.0\n'
        'Content-Transfer-Encoding: base64\n\n'
        'VGhpcyBlLW3DpGlsIGlzIGp1c3QgYSB0ZXN0\n'
    )


def test_send_email_unicode(smtp):
    app = Framework()
    app.mail_host, app.mail_port = smtp.address
    app.mail_sender = 'noreply@example.org'
    app.mail_force_tls = False
    app.mail_username = None
    app.mail_password = None
    app.mail_use_directory = False

    app.send_email(
        reply_to='👴 <info@example.org>',
        receivers=['recipient@example.org'],
        subject="👍",
        content="👍"
    )

    assert len(smtp.outbox) == 0
    transaction.commit()

    assert len(smtp.outbox) == 1
    message = smtp.outbox[0]

    def decode(header):
        name, addr = parseaddr(header)

        try:
            name = decode_header(name)[0][0].decode('utf-8')
        except AttributeError:
            pass

        return name, addr

    assert decode(message['Sender']) == ("👴", 'noreply@example.org')
    assert decode(message['From']) == ("👴", 'noreply@example.org')
    assert decode(message['Reply-To']) == ("👴", 'info@example.org')
    assert decode_header(message['Subject'])[0][0].decode('utf-8') == "👍"

    assert message.get_payload()[0].as_string() == (
        'Content-Type: text/plain; charset="utf-8"\n'
        'MIME-Version: 1.0\n'
        'Content-Transfer-Encoding: base64\n\n'
        '8J+RjQ==\n'
    )

    assert message.get_payload()[1].as_string() == (
        'Content-Type: text/html; charset="utf-8"\n'
        'MIME-Version: 1.0\n'
        'Content-Transfer-Encoding: base64\n\n'
        '8J+RjQ==\n'
    )


def test_convert_to_plaintext():
    html = (
        "<h1>Date</h1><p>6. April 1984</p>\n"
        "<h1>Path</h1><p>/foo-bar</p>\n"
    )

    plaintext = convert_to_plaintext(html)
    assert plaintext == "# Date\n\n6. April 1984\n\n# Path\n\n/foo-bar"


def test_object_by_path():

    class App(Framework):
        pass

    @App.path(path='/')
    class Root(object):
        pass

    @App.path(path='/pages', absorb=True)
    class Page(object):
        def __init__(self, absorb):
            self.absorb = absorb

    app = App()
    assert isinstance(app.object_by_path('/'), Root)
    assert isinstance(app.object_by_path('https://www.example.org/'), Root)

    page = app.object_by_path('/pages/foo/bar')
    assert page.absorb == 'foo/bar'
    assert isinstance(page, Page)

    assert app.object_by_path('/pages').absorb == ''

    # works, because 'foobar' is a view of the root
    assert isinstance(app.object_by_path('/foobar'), Root)
    assert app.object_by_path('/asdf/asdf') is None


def test_send_email_transaction(smtp):

    import more.transaction
    import more.webassets
    import onegov.core

    morepath.scan(more.transaction)
    morepath.scan(more.webassets)
    morepath.scan(onegov.core)

    class App(Framework):
        pass

    @App.path(path='/')
    class Root(object):
        pass

    @App.view(model=Root, name='send-fail')
    def fail_send(self, request):
        app.send_email(
            reply_to='Gövikon <info@example.org>',
            receivers=['recipient@example.org'],
            subject="Nüws",
            content="This e-mäil is just a test"
        )
        assert False

    @App.view(model=Root, name='send-ok')
    def success_send(self, request):
        app.send_email(
            reply_to='Gövikon <info@example.org>',
            receivers=['recipient@example.org'],
            subject="Nüws",
            content="This e-mäil is just a test"
        )

    app = App()
    app.application_id = 'test'
    app.configure_application(identity_secure=False)  # allow http
    app.mail_host, app.mail_port = smtp.address
    app.mail_sender = 'noreply@example.org'
    app.mail_force_tls = False
    app.mail_username = None
    app.mail_password = None
    app.mail_use_directory = False

    app.cache_backend = 'dogpile.cache.memory'
    app.cache_backend_arguments = {}

    client = Client(app)

    with pytest.raises(AssertionError):
        client.get('/send-fail')

    assert len(smtp.outbox) == 0

    client.get('/send-ok')
    assert len(smtp.outbox) == 1


def test_send_email_plaintext_alternative(smtp):
    app = Framework()
    app.mail_host, app.mail_port = smtp.address
    app.mail_sender = 'noreply@example.org'
    app.mail_force_tls = False
    app.mail_username = None
    app.mail_password = None
    app.mail_use_directory = False

    app.send_email(
        reply_to='Govikon <info@example.org>',
        receivers=['recipient@example.org'],
        subject="Test E-Mail",
        content="<a href='http://example.org'>This e-mail is just a test</a>"
    )

    transaction.commit()

    assert len(smtp.outbox) == 1
    message = smtp.outbox[0]

    assert message['Sender'] == 'Govikon <noreply@example.org>'
    assert message['From'] == 'Govikon <noreply@example.org>'
    assert message['Reply-To'] == 'Govikon <info@example.org>'
    assert message['Subject'] == 'Test E-Mail'

    assert message.get_payload()[0].as_string() == (
        'Content-Type: text/plain; charset="utf-8"\n'
        'MIME-Version: 1.0\n'
        'Content-Transfer-Encoding: base64\n\n'
        'W1RoaXMgZS1tYWlsIGlzIGp1c3QgYSB0ZXN0XShodHRwOi8vZXhhbXBsZS5vcmcp\n'
    )

    content = message.get_payload()[0].as_string().splitlines()[-1]
    assert b64decode(content) == (
        b"[This e-mail is just a test](http://example.org)"
    )

    assert message.get_payload()[1].as_string() == (
        'Content-Type: text/html; charset="utf-8"\n'
        'MIME-Version: 1.0\n'
        'Content-Transfer-Encoding: base64\n\n'
        'PGEgaHJlZj0naHR0cDovL2V4YW1wbGUub3JnJz5UaGlzI'
        'GUtbWFpbCBpcyBqdXN0IGEgdGVzdDwv\nYT4=\n'
    )

    content = ''.join(message.get_payload()[1].as_string().splitlines()[-2:])
    assert b64decode(content) == (
        b"<a href='http://example.org'>This e-mail is just a test</a>"
    )



class CSVError(Exception):
    pass


class MissingColumnsError(CSVError):
    def __init__(self, columns):
        self.columns = columns


class AmbiguousColumnsError(CSVError):
    def __init__(self, columns):
        self.columns = columns


class DuplicateColumnNamesError(CSVError):
    pass


class InvalidFormatError(CSVError):
    pass


class EmptyFileError(CSVError):
    pass


class EmptyLineInFileError(CSVError):
    pass


class AlreadyLockedError(Exception):
    """ Raised if :func:`onegov.core.locking.lock` fails to acquire a lock. """



import pytest

from onegov.core.crypto import (
    hash_password,
    random_token,
    verify_password,
    random_password
)


def test_hash_password():
    # because we use random salts we won't get the same result twice
    assert hash_password('hunter2') != hash_password('hunter2')

    # but that doesn't matter
    assert verify_password('hunter2', hash_password('hunter2'))
    assert verify_password('hunter2', hash_password('hunter2'))

    assert verify_password('hunter2', (
        '$bcrypt-sha256$2a,12$noO5p60IvoXlJN19'
        'lNwCQ.$sSGl3O6lQIdS8wFX/.i3NVc2HwNn5/.'
    ))


def test_random_token():
    assert random_token() != random_token()
    assert len(random_token()) == 64
    assert len(random_token(nbytes=1024)) == 64

    with pytest.raises(AssertionError):
        random_token(nbytes=511)


def test_no_null_bytes():
    with pytest.raises(AssertionError):
        assert hash_password("\0")

    with pytest.raises(AssertionError):
        assert verify_password("\0", "hash")


def test_random_password():
    assert len(random_password(16)) == 16
    assert len(random_password(32)) == 32



""" Provides a distributed lock implementation using Postgres Advisory Locks
as a backend.

"""

import contextlib
import zlib

from onegov.core.errors import AlreadyLockedError
from sqlalchemy import func


# Keeps the used locks around to make sure we detect collisions
USED_LOCKS = {}


@contextlib.contextmanager
def lock(session, namespace, key):
    """ A context manager which creates a distributed lock on the database.

    This is used to make sure that only one process at a time does something.
    This is not limited to database work (in fact this would usually be
    handled automatically). For example if multiple cronjobs are scheduled
    to run at the same time, but you only want one to actually run on
    the system on any given time, then you may use this lock.

    The locks are bound to the session and are not relased automatically
    if the current transaction is commited or rolled back. You don't have
    to worry about that though, the context manager releases all locks
    it creates once it's done.

    :session:
        The session to use for locking. The same lock may be acquired multiple
        times by the same session.

    :namespace:
        The namespace of the lock (any string).

    :key:
        The key of the lock (any string). The namespace combined with the
        key is used as the unique identifier which identifies the acquired
        lock.

    If the lock cannot be acquired the contextmanager raises a
    :class:`~onegov.core.errors.AlreadyLockedError`.

    By deafult sessions in onegov.core are bound to the thread as well as
    the current database schema. Therefore the lock is bound to those too.

    This is not a problem in practice however, as the request is bound the
    same way. At worst this makes it harder to share the lock, which is not a
    feature we are interested in. We want a guarantee that only one process or
    thread at any given time is doing something using a lock. This works.

    Example:

        with lock(session, 'cronjobs', 'send-email'):
            pass  # send email

    """
    id = get_lock_id(namespace, key)

    locked = session.execute(func.pg_try_advisory_lock(id)).scalar()

    if not locked:
        raise AlreadyLockedError()

    try:
        yield
    finally:
        session.execute(func.pg_advisory_unlock(id))


def get_lock_id(namespace, key):
    """ Returns a lock id for the given namespace and key. The lock id is
    an integer between -2**63 - 1 and 2**63 - 1 (signed 64bit int).

    Because that leaves us with a rather small space to generate a unique
    lock id from namespace and key we use two crc32 bit hashes and combine
    them.

    This still leaves us with a chance of duplicate locks. Since those strings
    are static at this point we keep a list of existing locks around to
    catch any such case during testing. If that ever happens we need to change
    our strings.

    Note: We can't just use an incrementing counter because those strings
    have to match up over multiple processes which wouldn't necessarily
    count up in the same order.

    """

    namespace = namespace.encode('utf-8')
    key = key.encode('utf-8')

    lock_id = (32 >> zlib.crc32(namespace)) | zlib.crc32(key)

    combination = (namespace, key)

    # If this assertion occurs we have a namespace collision between two
    # different namespace/key pairs. This might never happen. If it does
    # see the documentation of the current function.
    assert USED_LOCKS.get(lock_id, combination) == combination
    USED_LOCKS[lock_id] = combination

    return lock_id



import os
import os.path
import polib
import pytest

from gettext import GNUTranslations, NullTranslations
from onegov.core import i18n, utils
from webob import Request
from webob.acceptparse import Accept
from wtforms import Label


def test_pofiles(temporary_directory):
    os.makedirs(os.path.join(temporary_directory, 'de/LC_MESSAGES'))
    utils.touch(
        os.path.join(temporary_directory, 'de/LC_MESSAGES/onegov.test.po'))

    result = list(i18n.pofiles(temporary_directory))
    assert result[0][0] == 'de'
    assert result[0][1].endswith('onegov.test.po')


def test_merge_translations():
    t1 = GNUTranslations()
    t1._catalog = {
        'We are here': 'Hier sind wir',
        'Welcome': 'Willkommen'
    }

    t2 = GNUTranslations()
    t2._catalog = {
        'We are here': 'Wir sind hier',
        'Hi': 'Hallo'
    }

    t3 = GNUTranslations()
    t3._catalog = {
        'We are here': 'Sind wir hier',
        'Morning': 'Moin'
    }

    with pytest.raises(AssertionError):
        i18n.merge([])

    with pytest.raises(AssertionError):
        i18n.merge([i18n.clone(t1)])

    assert i18n.merge([i18n.clone(t1), i18n.clone(t2)])._catalog == {
        'We are here': 'Hier sind wir',
        'Welcome': 'Willkommen',
        'Hi': 'Hallo'
    }

    assert i18n.merge([i18n.clone(t2), i18n.clone(t3)])._catalog == {
        'We are here': 'Wir sind hier',
        'Morning': 'Moin',
        'Hi': 'Hallo'
    }

    assert i18n.merge(
        [i18n.clone(t1), i18n.clone(t2), i18n.clone(t3)])._catalog == {
        'We are here': 'Hier sind wir',
        'Welcome': 'Willkommen',
        'Morning': 'Moin',
        'Hi': 'Hallo'
    }


def test_get_translations(temporary_directory):
    os.makedirs(os.path.join(temporary_directory, '1/de/LC_MESSAGES'))
    po = polib.POFile()
    po.append(polib.POEntry(
        msgid='Welcome',
        msgstr='Willkommen'
    ))
    po.save(os.path.join(
        temporary_directory, '1/de/LC_MESSAGES/onegov.test.po'))

    os.makedirs(os.path.join(temporary_directory, '2/de/LC_MESSAGES'))
    po = polib.POFile()
    po.append(polib.POEntry(
        msgid='Morning',
        msgstr='Moin'
    ))
    po.save(os.path.join(
        temporary_directory, '2/de/LC_MESSAGES/onegov.test.po'))

    os.makedirs(os.path.join(temporary_directory, '2/fr/LC_MESSAGES'))
    po = polib.POFile()
    po.append(polib.POEntry(
        msgid='Welcome',
        msgstr='Bienvenue'
    ))
    po.save(os.path.join(
        temporary_directory, '2/fr/LC_MESSAGES/onegov.test.po'))

    translations = i18n.get_translations([
        temporary_directory + '/1',
        temporary_directory + '/2'
    ])
    assert list(sorted(translations.keys())) == list(sorted(['de', 'fr']))

    translations['de'].gettext('Welcome') == 'Willkommen'
    translations['de'].gettext('Morning') == 'Moin'
    translations['fr'].gettext('Welcome') == 'Bienvenue'

    translations = i18n.get_translations([
        temporary_directory + '/1',
    ])
    assert list(sorted(translations.keys())) == list(sorted(['de']))
    translations['de'].gettext('Welcome') == 'Willkommen'
    translations['de'].gettext('Morning') == 'Morning'

    translations = i18n.get_translations([
        temporary_directory + '/2',
    ])
    assert list(sorted(translations.keys())) == list(sorted(['de', 'fr']))
    translations['de'].gettext('Welcome') == 'Welcome'
    translations['de'].gettext('Morning') == 'Moin'
    translations['fr'].gettext('Welcome') == 'Bienvenue'


def test_default_locale_negotiator():
    negotiate = i18n.default_locale_negotiator
    request = Request.blank('/')

    request.accept_language = Accept('fr')
    assert negotiate(['en', 'de'], request) is None

    request.accept_language = Accept('en')
    assert negotiate(['en', 'de'], request) == 'en'

    request.accept_language = Accept('de')
    assert negotiate(['en', 'de'], request) == 'de'

    request.accept_language = Accept('de')
    request.cookies['locale'] = 'en'
    assert negotiate(['en', 'de'], request) == 'en'


def test_add_fallback_to_tail():

    root = NullTranslations()
    child = NullTranslations()
    grandchild = NullTranslations()

    i18n.add_fallback_to_tail(root, child)

    assert root._fallback is child
    assert child._fallback is None

    i18n.add_fallback_to_tail(root, grandchild)

    assert root._fallback is child
    assert child._fallback is grandchild
    assert grandchild._fallback is None
    assert root._fallback._fallback is grandchild

    i18n.add_fallback_to_tail(root, child)

    assert root._fallback is child
    assert child._fallback is grandchild
    assert grandchild._fallback is None
    assert root._fallback._fallback is grandchild

    i18n.add_fallback_to_tail(root, grandchild)

    assert root._fallback is child
    assert child._fallback is grandchild
    assert grandchild._fallback is None
    assert root._fallback._fallback is grandchild


def test_get_translation_bound_form():

    class MockTranslation(object):

        _fallback = None

        def add_fallback(self, translation):
            self._fallback = translation

        def gettext(self, text):
            return text[::-1]  # inverse the string

    default = MockTranslation()
    translate = MockTranslation()

    class MetaMeta(object):

        def render_field(self, field, render_kw):
            return field.label.text

    class MockMeta(MetaMeta):

        def get_translations(self, form):
            return default

    class MockForm(object):

        Meta = MockMeta

    class MockField(object):

        def __init__(self, label):
            self.label = Label('x', label)

    form_class = i18n.get_translation_bound_form(MockForm, translate)

    assert form_class.Meta().get_translations(None) is default
    assert form_class.Meta().get_translations(None)._fallback is translate

    meta = form_class.Meta()
    meta.get_translations(None)
    assert meta.render_field(MockField('Yes'), {}) == 'seY'



import morepath
import os
import os.path
import polib

from onegov.core import Framework
from onegov.core import utils
from onegov.core.layout import ChameleonLayout
from onegov.core.templates import render_macro
from translationstring import TranslationStringFactory
from webtest import TestApp as Client


def test_chameleon_with_translation(temporary_directory):
    # Test chameleon in a real scenario with templating and translations

    templates = os.path.join(temporary_directory, 'templates')
    os.mkdir(templates)

    locale = os.path.join(temporary_directory, 'locale')
    os.makedirs(os.path.join(locale, 'de/LC_MESSAGES'))

    po = polib.POFile()
    po.append(polib.POEntry(
        msgid='Welcome',
        msgstr='Willkommen'
    ))
    po.append(polib.POEntry(
        msgid="We're sinking, we're sinking!",
        msgstr='Wir denken, wir denken!'
    ))
    po.append(polib.POEntry(
        msgid="Macro",
        msgstr="Makro"
    ))
    po.save(os.path.join(locale, 'de/LC_MESSAGES/onegov.test.po'))

    with open(os.path.join(templates, 'index.pt'), 'w') as f:
        f.write("""
            <!DOCTYPE html>
            <html xmlns="http://www.w3.org/1999/xhtml"
                  xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                  xmlns:tal="http://xml.zope.org/namespaces/tal"
                  i18n:domain="onegov.test">
                <b><tal:block i18n:translate="">Welcome</tal:block></b>
                <b>${foo}</b>
                <metal:macro
                    define-macro="testmacro" i18n:domain="onegov.test">
                    <i i18n:translate>Macro</i>
                </metal:macro>
            </html>
        """)

    _ = TranslationStringFactory('onegov.test')

    class App(Framework):
        pass

    @App.template_directory()
    def get_template_directory():
        return templates

    @App.setting(section='i18n', name='localedirs')
    def get_localedirs():
        return [locale]

    @App.setting(section='i18n', name='default_locale')
    def get_i18n_default_locale():
        return 'de'

    @App.path(path='/')
    class Root(object):
        pass

    @App.html(model=Root, template='index.pt')
    def view_root(self, request):
        return {
            'foo': _("We're sinking, we're sinking!")
        }

    @App.view(model=Root, name='macro')
    def view_root_macro(self, request):
        registry = request.app.config.template_engine_registry
        template = registry._template_loaders['.pt']['index.pt']
        template.macros['testmacro']

        return render_macro(template.macros['testmacro'], request, {})

    utils.scan_morepath_modules(App)

    client = Client(App())
    assert '<b>Willkommen</b>' in client.get('/').text
    assert '<b>Wir denken, wir denken!</b>' in client.get('/').text
    assert '<i>Makro</i>' in client.get('/').text

    assert '<b>Willkommen</b>' not in client.get('/macro').text
    assert '<b>Wir denken, wir denken!</b>' not in client.get('/macro').text

    assert '<i>Makro</i>' in client.get('/macro').text


def test_inject_default_vars(temporary_directory):

    templates = os.path.join(temporary_directory, 'templates')
    os.mkdir(templates)

    with open(os.path.join(templates, 'index.pt'), 'w') as f:
        f.write("""
            <!DOCTYPE html>
            <html xmlns="http://www.w3.org/1999/xhtml"
                  xmlns:tal="http://xml.zope.org/namespaces/tal">
                <span>${injected}</span>
                <span tal:condition="parent|nothing">padre</span>
                <span tal:condition="child|nothing">niño</span>
            </html>
        """)

    class Parent(Framework):
        pass

    @Parent.template_directory()
    def get_template_directory():
        return templates

    @Parent.path(path='/')
    class Root(object):
        pass

    @Parent.view(model=Root, template='index.pt')
    def view_root(self, request):
        return {
            'injected': 'foobar'
        }

    @Parent.template_variables()
    def get_parent_template_variables(request):
        return {
            'injected': 'parent',
            'parent': True
        }

    class Child(Parent):
        pass

    @Child.template_variables()
    def get_child_template_variables(request):
        return {
            'injected': 'child',
            'child': True
        }

    utils.scan_morepath_modules(Parent)
    utils.scan_morepath_modules(Child)

    morepath.commit(Child, Parent)

    parent_page = Client(Parent()).get('/')
    assert 'parent' in parent_page
    assert 'child' not in parent_page
    assert 'padre' in parent_page
    assert 'niño' not in parent_page

    child_page = Client(Child()).get('/')
    assert 'parent' not in child_page
    assert 'child' in child_page
    assert 'padre' in child_page
    assert 'niño' in child_page


def test_macro_lookup(temporary_directory):

    parent = os.path.join(temporary_directory, 'parent')
    child = os.path.join(temporary_directory, 'child')

    os.mkdir(parent)
    os.mkdir(child)

    with open(os.path.join(parent, 'index.pt'), 'w') as f:
        f.write("""
            <!DOCTYPE html>
            <html xmlns="http://www.w3.org/1999/xhtml"
                  xmlns:tal="http://xml.zope.org/namespaces/tal"
                  xmlns:metal="http://xml.zope.org/namespaces/metal">
                <metal:block use-macro="layout.macros.foo" />
                <metal:block use-macro="layout.macros.bar" />
                <metal:block use-macro="layout.macros.id" />
            </html>
        """)

    with open(os.path.join(parent, 'macros.pt'), 'w') as f:
        f.write("""
            <!DOCTYPE html>
            <html xmlns="http://www.w3.org/1999/xhtml"
                  xmlns:tal="http://xml.zope.org/namespaces/tal"
                  xmlns:metal="http://xml.zope.org/namespaces/metal">

                <metal:block define-macro="foo">
                    Foo
                </metal:block>

                <metal:block define-macro="id">
                    Parent
                </metal:block>
            </html>
        """)

    with open(os.path.join(child, 'macros.pt'), 'w') as f:
        f.write("""
            <!DOCTYPE html>
            <html xmlns="http://www.w3.org/1999/xhtml"
                  xmlns:tal="http://xml.zope.org/namespaces/tal"
                  xmlns:metal="http://xml.zope.org/namespaces/metal">

                <metal:block define-macro="bar">
                    Bar
                </metal:block>

                <metal:block define-macro="id">
                    Child
                </metal:block>
            </html>
        """)

    class Parent(Framework):
        pass

    @Parent.template_directory()
    def get_parent_template_directory():
        return parent

    @Parent.path(path='/')
    class Root(object):
        pass

    @Parent.view(model=Root, template='index.pt')
    def view_root(self, request):
        return {
            'layout': ChameleonLayout(self, request)
        }

    class Child(Parent):
        pass

    @Child.template_directory()
    def get_child_template_directory():
        return child

    utils.scan_morepath_modules(Parent)
    utils.scan_morepath_modules(Child)

    morepath.commit(Child, Parent)

    page = Client(Child()).get('/')
    assert 'Foo' in page
    assert 'Bar' in page
    assert 'Child' in page



import os.path

from datetime import datetime
from dectate import Action
from inspect import isfunction
from morepath.directive import HtmlAction
from onegov.core.cronjobs import register_cronjob
from onegov.core.framework import Framework
from onegov.core.utils import Bunch
from sedate import to_timezone, replace_timezone


@Framework.directive('form')
class HtmlHandleFormAction(HtmlAction):
    """ Register Form view.

    Basically wraps the Morepath's ``html`` directive, registering both
    POST and GET (if no specific request method is given) and wrapping the
    view handler with :func:`wrap_with_generic_form_handler`.

    The form is either a class or a function. If it's a function, it is
    expected to return a form class when given an instance of the model.

    The form may also be None, which is useful under special circumstances.
    Generally you don't want that though.

    Example:

    .. code-block:: python

        @App.form(model=Root, template='form.pt',
                  permission=Public, form=LoginForm)
        def handle_form(self, request, form):
            if form.submitted():
                # do something if the form was submitted with valid data
            else:
                # do something if the form was not submitted or not
                # submitted correctly

            return {}  # template variables

    """
    def __init__(self, model, form, render=None, template=None,
                 permission=None, internal=False, **predicates):
        self.form = form
        super().__init__(model, render, template, permission, internal,
                         **predicates)

    def perform(self, obj, view_registry):

        keys = self.key_dict()

        wrapped = wrap_with_generic_form_handler(
            obj, self.form, keys.get('name'))

        if 'request_method' not in keys:

            keys['request_method'] = 'GET'
            view_registry.register_view(
                keys, wrapped,
                self.render, self.template,
                self.permission, self.internal)

            keys['request_method'] = 'POST'
            view_registry.register_view(
                keys, wrapped,
                self.render, self.template,
                self.permission, self.internal)
        else:
            view_registry.register_view(
                keys, wrapped,
                self.render, self.template,
                self.permission, self.internal)


def wrap_with_generic_form_handler(obj, form_class, view_name):
    """ Wraps a view handler with generic form handling.

    This includes instantiatng the form with translations/csrf protection
    and setting the correct action.

    """

    def handle_form(self, request):

        if isfunction(form_class):
            _class = form_class(self, request)
        else:
            _class = form_class

        if _class:
            form = request.get_form(_class)
            form.action = request.link(self, name=view_name)
        else:
            form = None

        return obj(self, request, form)

    return handle_form


@Framework.directive('cronjob')
class CronjobAction(Action):
    """ Register a cronjob. """

    config = {
        'cronjob_registry': Bunch
    }

    def __init__(self, hour, minute, timezone):
        self.hour = hour
        self.minute = minute
        self.timezone = timezone

    def identifier(self, cronjob_registry):
        # return a key to the hour/minute on a fixed day, this way it's
        # impossible to have two cronjobs running at the exact same time.
        return to_timezone(
            replace_timezone(
                datetime(2016, 1, 1, self.hour, self.minute),
                self.timezone
            ), 'UTC'
        )

    def perform(self, func, cronjob_registry):
        register_cronjob(
            cronjob_registry, func, self.hour, self.minute, self.timezone)


@Framework.directive('static_directory')
class StaticDirectoryAction(Action):
    """ Registers a static files directory. """

    config = {
        'staticdirectory_registry': Bunch
    }

    counter = iter(range(1, 123456789))

    def __init__(self):
        self.name = next(self.counter)

    def identifier(self, staticdirectory_registry):
        return self.name

    def perform(self, func, staticdirectory_registry):
        if not hasattr(staticdirectory_registry, 'paths'):
            staticdirectory_registry.paths = []

        path = func()

        if not os.path.isabs(path):
            path = os.path.join(os.path.dirname(self.code_info.path), path)

        staticdirectory_registry.paths.append(path)


class TemplateVariablesRegistry(object):

    __slots__ = ['callbacks']

    def __init__(self):
        self.callbacks = []

    def get_variables(self, request, base=None):
        base = base or {}

        for callback in self.callbacks:
            base.update(callback(request))

        return base


@Framework.directive('template_variables')
class TemplateVariablesAction(Action):
    """ Registers a set of global template variables for chameleon templates.

    Only exists once per application. Template variables with conflicting
    keys defined in child applications override the keys with the same
    name in the parent application. Non-conflicting keys are kept individually.

    Example::

        @App.template_variables()
        def get_template_variables(request):
            return {
                'foo': 'bar'
            }

    """

    config = {
        'templatevariables_registry': TemplateVariablesRegistry
    }

    counter = iter(range(1, 123456789))

    def __init__(self):
        # XXX I would expect this to work with a static name (and it does in
        # tests), but in real world usage the same name leads to overriden
        # paths
        self.name = next(self.counter)

    def identifier(self, templatevariables_registry):
        return self.name

    def perform(self, func, templatevariables_registry):
        templatevariables_registry.callbacks.append(func)



import pytest

from dogpile.cache import make_region
from dogpile.cache.api import NO_VALUE
from onegov.core.browser_session import BrowserSession


def test_browser_session_mangle():
    session = BrowserSession('ns', 'token', {})
    assert session.mangle('test') == 'ns:token:test'

    with pytest.raises(AssertionError):
        session.mangle('')

    with pytest.raises(AssertionError):
        session.mangle(None)


def test_browser_session_cache():
    cache = make_region().configure('dogpile.cache.memory')
    session = BrowserSession('ns', 'token', cache)

    assert not session.has('name')
    assert 'name' not in session

    with pytest.raises(AttributeError):
        session.name

    with pytest.raises(KeyError):
        session['name']

    assert cache.get('ns:token:name') is NO_VALUE

    session.name = 'test'
    assert session.has('name')
    assert 'name' in session
    assert session.name == 'test'
    assert session['name'] == 'test'
    assert cache.get('ns:token:name') == 'test'

    del session.name
    assert not session.has('name')
    assert 'name' not in session

    with pytest.raises(AttributeError):
        session.name

    with pytest.raises(KeyError):
        session['name']

    assert cache.get('ns:token:name') is NO_VALUE


def test_browser_session_cache_namespacing():
    cache = make_region().configure('dogpile.cache.memory')

    session = BrowserSession('ns', 'token', cache)
    session.name = 'test'
    assert session.name == 'test'

    session = BrowserSession('ns', 'token', cache)
    assert session.name == 'test'

    session = BrowserSession('ns', 'token2', cache)
    with pytest.raises(AttributeError):
        session.name

    session = BrowserSession('ns2', 'token', cache)
    with pytest.raises(AttributeError):
        session.name



""" Contains custom converters used by onegov.town. """

import morepath

from datetime import date
from onegov.core import Framework
from onegov.core.utils import is_uuid
from time import mktime, strptime
from uuid import UUID


def extended_date_decode(s):
    """ Decodes a date string HTML5 (RFC3339) compliant."""
    if not s:
        return None

    return date.fromtimestamp(mktime(strptime(s, '%Y-%m-%d')))


def extended_date_encode(d):
    """ Encodes a date HTML5 (RFC3339) compliant. """
    if not d:
        return ''

    return d.strftime('%Y-%m-%d')


extended_date_converter = morepath.Converter(
    decode=extended_date_decode, encode=extended_date_encode
)


def uuid_decode(s):
    """ Turns a uuid string into a UUID instance. """

    return is_uuid(s) and UUID(s) or None


def uuid_encode(uuid):
    """ Turns a UUID instance into a uuid string. """
    if not uuid:
        return ''

    if isinstance(uuid, str):
        return uuid

    return uuid.hex


uuid_converter = morepath.Converter(
    decode=uuid_decode, encode=uuid_encode
)


@Framework.converter(type=UUID)
def get_default_uuid_converter():
    return uuid_converter
